import importlib
import json
import logging
import os
import re
from pathlib import Path
from tempfile import mkstemp
from urllib.parse import urlparse

import pypandoc
from django.apps import apps
from django.conf import settings
from django.http import Http404, HttpResponse, HttpResponseBadRequest
from django.template.loader import get_template
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from markdown import markdown

from defusedcsv import csv

log = logging.getLogger(__name__)


def get_script_alias(request):
    return request.path[:-len(request.path_info)]


def get_referer(request, default=None):
    return request.META.get('HTTP_REFERER', default)


def get_referer_path_info(request, default=''):
    referer = request.META.get('HTTP_REFERER', None)
    if not referer:
        return default

    script_alias = get_script_alias(request)
    return urlparse(referer).path[len(script_alias):]


def get_next(request):
    next = request.POST.get('next')
    current = request.path_info

    if next in (current, None):
        return get_script_alias(request) + '/'
    else:
        return get_script_alias(request) + next


def get_uri_prefix(obj):
    # needs to stay, is part of a migration
    r = settings.DEFAULT_URI_PREFIX
    if bool(obj.uri_prefix) is True:
        r = obj.uri_prefix.rstrip('/')
    return r


def get_pandoc_main_version():
    return int(pypandoc.get_pandoc_version().split('.')[0])


def pandoc_version_at_least(required_version):
    required = [int(x) for x in required_version.split('.')]
    installed = [int(x) for x in pypandoc.get_pandoc_version().split('.')]
    for idx, digit in enumerate(installed):
        try:
            req = required[idx]
        except IndexError:
            return True
        else:
            if digit < req:
                return False
            if digit > req:
                return True
    return True


def join_url(base, *args):
    url = base
    for arg in args:
        url = url.rstrip('/') + '/' + arg.lstrip('/')
    return url


def get_model_field_meta(model):
    meta = {}

    for field in model._meta.get_fields():
        meta[field.name] = {}
        if hasattr(field, 'verbose_name'):
            meta[field.name]['verbose_name'] = field.verbose_name
        if hasattr(field, 'help_text'):
            meta[field.name]['help_text'] = field.help_text

    return meta


def get_languages():
    languages = []
    for i in range(5):
        try:
            language = settings.LANGUAGES[i][0], settings.LANGUAGES[i][1],\
                'lang%i' % (i + 1)
            languages.append(language)
        except IndexError:
            pass
    return languages


def get_language_fields(field_name):
    return [
        field_name + '_' + lang_field for lang_code,
        lang_string, lang_field in get_languages()
        ]


def get_language_warning(obj, field):
    for lang_code, lang_string, lang_field in get_languages():
        if not getattr(obj, '%s_%s' % (field, lang_field)):
            return True
    return False


def set_export_reference_document(format, context):
    # try to get the view uri from the context
    try:
        view = context['view']
        view_uri = getattr(view, 'uri')
    except (AttributeError, KeyError, TypeError):
        view_uri = None

    refdocs = []

    if format == 'odt':
        # append view specific custom refdoc
        try:
            refdocs.append(settings.EXPORT_REFERENCE_ODT_VIEWS[view_uri])
        except KeyError:
            pass

        # append custom refdoc
        if settings.EXPORT_REFERENCE_ODT:
            refdocs.append(settings.EXPORT_REFERENCE_ODT)

    elif format == 'docx':
        # append view specific custom refdoc
        try:
            refdocs.append(settings.EXPORT_REFERENCE_DOCX_VIEWS[view_uri])
        except KeyError:
            pass

        # append custom refdoc
        if settings.EXPORT_REFERENCE_DOCX:
            refdocs.append(settings.EXPORT_REFERENCE_DOCX)

    # append the default reference docs
    refdocs.append(
        os.path.join(
            apps.get_app_config('rdmo').path,
            'share', 'reference' + '.' + format
        )
    )

    # return the first file in refdocs that actually exists
    for refdoc in refdocs:
        if os.path.isfile(refdoc):
            return refdoc


def render_to_format(request, export_format, title, template_src, context):
    if export_format not in dict(settings.EXPORT_FORMATS):
        return HttpResponseBadRequest(_('This format is not supported.'))

    # render the template to a html string
    template = get_template(template_src)
    html = template.render(context)
    metadata, html = parse_metadata(html)

    # remove empty lines
    html = os.linesep.join([line for line in html.splitlines() if line.strip()])

    if export_format == 'html':
        # create the response object
        response = HttpResponse(html)

    else:
        pandoc_args = settings.EXPORT_PANDOC_ARGS.get(export_format, [])
        content_disposition = 'attachment; filename="%s.%s"' % (title, export_format)

        if export_format == 'pdf':
            # check pandoc version (the pdf arg changed to version 2)
            if get_pandoc_main_version() == 1:
                pandoc_args = [arg.replace(
                    '--pdf-engine=xelatex', '--latex-engine=xelatex'
                ) for arg in pandoc_args]

            # display pdf in browser
            content_disposition = 'filename="%s.%s"' % (title, export_format)

        # use reference document for certain file formats
        refdoc = set_export_reference_document(export_format, context)
        if refdoc is not None and export_format in ['docx', 'odt']:
            # check pandoc version (the args changed to version 2)
            if get_pandoc_main_version() == 1:
                pandoc_args.append('--reference-{}={}'.format(export_format, refdoc))
            else:
                pandoc_args.append('--reference-doc={}'.format(refdoc))

        # add the possible resource-path
        if pandoc_version_at_least("2") is True:
            pandoc_args.append('--resource-path={}'.format(settings.STATIC_ROOT))
            if 'resource_path' in context:
                resource_path = Path(settings.MEDIA_ROOT).joinpath(context['resource_path'])
                pandoc_args.append('--resource-path={}'.format(resource_path))

        # create a temporary file
        (tmp_fd, tmp_filename) = mkstemp('.' + export_format)

        # add metadata
        tmp_metadata_file = None
        if metadata is not None and pandoc_version_at_least("2.3") is True:
            tmp_metadata_file = save_metadata(metadata)
            pandoc_args.append('--metadata-file=' + tmp_metadata_file)

        # convert the file using pandoc
        log.info('Export %s document using args %s.', export_format, pandoc_args)
        html = re.sub(
            r'(<img.+src=["\'])' + settings.STATIC_URL + r'([\w\-\@?^=%&/~\+#]+)', r'\g<1>' +
            str(Path(settings.STATIC_ROOT)) + r'/\g<2>', html
        )
        pypandoc.convert_text(
            html, export_format, format='html',
            outputfile=tmp_filename, extra_args=pandoc_args
        )

        # read the temporary file
        file_handler = os.fdopen(tmp_fd, 'rb')
        file_content = file_handler.read()
        file_handler.close()

        # delete temporary files
        if tmp_metadata_file is not None:
            os.remove(tmp_metadata_file)
        os.remove(tmp_filename)

        # create the response object
        response = HttpResponse(file_content, content_type='application/%s' % export_format)
        response['Content-Disposition'] = content_disposition.encode('utf-8')

    return response


def render_to_csv(title, rows, delimiter=','):
    response = HttpResponse(content_type='text/csv')
    response['Content-Disposition'] = 'attachment; filename="%s.csv"' % title

    writer = csv.writer(response, delimiter=delimiter)
    for row in rows:
        writer.writerow(
            ['' if x is None else str(x) for x in row]
        )
    return response


def return_file_response(file_path, content_type):
    file_abspath = Path(settings.MEDIA_ROOT) / file_path
    if file_abspath.exists():
        with file_abspath.open('rb') as fp:
            response = HttpResponse(fp.read(), content_type=content_type)
            response['Content-Disposition'] = 'attachment; filename=' + file_abspath.name
            return response
    else:
        raise Http404


def sanitize_url(s):
    # is used in the rdmo-app
    try:
        m = re.search('[a-z0-9-_]', s)
    except TypeError:
        s = ''
    else:
        if bool(m) is False:
            s = ''
        else:
            s = re.sub(r'/+', '/', s)
    return s


def import_class(string):
    module_name, class_name = string.rsplit('.', 1)
    return getattr(importlib.import_module(module_name), class_name)


def copy_model(instance, **kwargs):
    # get values from instance which are not id, ForeignKeys orde M2M relations
    data = {}
    for field in instance._meta.get_fields():
        if not (field.name == 'id' or field.is_relation):
            data[field.name] = getattr(instance, field.name)

    # update with the kwargs provided to this function
    data.update(kwargs)

    # create and save new instance
    instance_copy = instance._meta.model(**data)
    instance_copy.save()

    return instance_copy


def human2bytes(string):
    if not string:
        return 0

    m = re.match(r'([0-9.]+)\s*([A-Za-z]+)', string)
    number, unit = float(m.group(1)), m.group(2).strip().lower()

    if unit == 'kb' or unit == 'k':
        return number * 1000
    elif unit == 'mb' or unit == 'm':
        return number * 1000**2
    elif unit == 'gb' or unit == 'g':
        return number * 1000**3
    elif unit == 'tb' or unit == 't':
        return number * 1000**4
    elif unit == 'pb' or unit == 'p':
        return number * 1000**5
    elif unit == 'kib':
        return number * 1024
    elif unit == 'mib':
        return number * 1024**2
    elif unit == 'gib':
        return number * 1024**3
    elif unit == 'tib':
        return number * 1024**4
    elif unit == 'pib':
        return number * 1024**5


def markdown2html(markdown_string):
    # adoption of the normal markdown function which also converts
    # `[<string>]{<title>}` to <span title="<title>"><string></span> to
    # allow for underlined tooltips
    html = markdown(force_str(markdown_string))
    html = re.sub(
        r'\[(.*?)\]\{(.*?)\}',
        '<span data-toggle="tooltip" data-placement="bottom" data-html="true" title="\2">\1</span>',
        html
    )
    return html


def parse_metadata(html):
    metadata = None
    pattern = re.compile(
        '(<metadata>)(.*)(</metadata>)', re.MULTILINE | re.DOTALL
    )
    m = re.search(pattern, html)
    if bool(m) is True:
        try:
            metadata = json.loads(m.group(2))
        except json.JSONDecodeError:
            pass
        else:
            html = html.replace(m.group(0), '')
    return metadata, html


def save_metadata(metadata):
    _, tmp_metadata_file = mkstemp(suffix='.json')
    with open(tmp_metadata_file, 'w') as f:
        json.dump(metadata, f)
    f = open(tmp_metadata_file)
    log.info('Save metadata file %s %s', tmp_metadata_file, str(metadata))
    return tmp_metadata_file



from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from rdmo.core.plugins import get_plugin

from ..managers import IntegrationManager


class Integration(models.Model):

    objects = IntegrationManager()

    project = models.ForeignKey(
        'Project', on_delete=models.CASCADE, related_name='integrations',
        verbose_name=_('Project'),
        help_text=_('The project for this integration.')
    )
    provider_key = models.TextField(
        verbose_name=_('Provider key'),
        help_text=_('The key of the provider for this integration.')
    )

    class Meta:
        ordering = ('project__title', )
        verbose_name = _('Integration')
        verbose_name_plural = _('Integrations')

    def __str__(self):
        return '%s / %s' % (self.project.title, self.provider_key)

    @property
    def provider(self):
        return get_plugin('SERVICE_PROVIDERS', self.provider_key)

    def get_absolute_url(self):
        return reverse('project', kwargs={'pk': self.project.pk})

    def save_options(self, options):
        for field in self.provider.fields:
            try:
                integration_option = IntegrationOption.objects.get(integration=self, key=field.get('key'))
            except IntegrationOption.DoesNotExist:
                integration_option = IntegrationOption(integration=self, key=field.get('key'))

            integration_option.value = options.get(field.get('key')) or ''  # required was (hopefully) checked before
            integration_option.secret = field.get('secret', False)
            integration_option.save()


class IntegrationOption(models.Model):

    integration = models.ForeignKey(
        'Integration', on_delete=models.CASCADE, related_name='options',
        verbose_name=_('Integration'),
        help_text=_('The integration for this integration option.')
    )
    key = models.SlugField(
        max_length=128,
        verbose_name=_('Key'),
        help_text=_('The key for this integration option.')
    )
    value = models.TextField(
        verbose_name=_('Value'),
        help_text=_('The value for this integration option.')
    )
    secret = models.BooleanField(
        default=False,
        verbose_name=_('Secret'),
        help_text=_('Designates whether this integration option is hidden from regular users.')
    )

    class Meta:
        ordering = ('integration__project__title', )
        verbose_name = _('Integration option')
        verbose_name_plural = _('Integration options')

    def __str__(self):
        return '%s / %s / %s = %s' % (self.integration.project.title, self.integration.provider_key, self.key, self.value)



import pytest
from django.core import mail
from django.urls import reverse

from rdmo.core.constants import VALUE_TYPE_FILE

from ..models import Issue, Project

users = (
    ('owner', 'owner'),
    ('manager', 'manager'),
    ('author', 'author'),
    ('guest', 'guest'),
    ('user', 'user'),
    ('site', 'site'),
    ('anonymous', None),
)

view_issue_permission_map = {
    'owner': [1, 2, 3, 4, 5],
    'manager': [1, 3, 5],
    'author': [1, 3, 5],
    'guest': [1, 3, 5],
    'api': [1, 2, 3, 4, 5],
    'site': [1, 2, 3, 4, 5]
}

change_issue_permission_map = {
    'owner': [1, 2, 3, 4, 5],
    'manager': [1, 3, 5],
    'author': [1, 3, 5],
    'api': [1, 2, 3, 4, 5],
    'site': [1, 2, 3, 4, 5]
}

delete_issue_permission_map = {
    'owner': [1, 2, 3, 4, 5],
    'manager': [1, 3, 5],
    'api': [1, 2, 3, 4, 5],
    'site': [1, 2, 3, 4, 5]
}


projects = [1, 2, 3, 4, 5]
issues = [1, 2, 3, 4]

integration_pk = 1

issue_status = ('open', 'in_progress', 'closed')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
@pytest.mark.parametrize('issue_id', issues)
def test_issue_update_get(db, client, username, password, project_id, issue_id):
    client.login(username=username, password=password)
    issue = Issue.objects.filter(project_id=project_id, id=issue_id).first()

    url = reverse('issue_update', args=[project_id, issue_id])
    response = client.get(url)

    if issue:
        if project_id in change_issue_permission_map.get(username, []):
            assert response.status_code == 200
        elif password:
            assert response.status_code == 403
        else:
            assert response.status_code == 302
    else:
        assert response.status_code == 404


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
@pytest.mark.parametrize('issue_id', issues)
@pytest.mark.parametrize('status', issue_status)
def test_issue_update_post(db, client, username, password, project_id, issue_id, status):
    client.login(username=username, password=password)
    issue = Issue.objects.filter(project_id=project_id, id=issue_id).first()

    url = reverse('issue_update', args=[project_id, issue_id])
    data = {
        'status': status
    }
    response = client.post(url, data)

    if issue:
        if project_id in change_issue_permission_map.get(username, []):
            assert response.status_code == 302
            assert Issue.objects.get(id=issue_id).status == status
        else:
            if password:
                assert response.status_code == 403
            else:
                assert response.status_code == 302

            assert Issue.objects.get(id=issue_id).status == issue.status
    else:
        assert response.status_code == 404


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('issue_id', issues)
@pytest.mark.parametrize('project_id', projects)
def test_issue_send_get(db, client, username, password, project_id, issue_id):
    client.login(username=username, password=password)
    issue = Issue.objects.filter(project_id=project_id, id=issue_id).first()

    url = reverse('issue_update', args=[project_id, issue_id])
    response = client.get(url)

    if issue:
        if project_id in change_issue_permission_map.get(username, []):
            assert response.status_code == 200
        elif password:
            assert response.status_code == 403
        else:
            assert response.status_code == 302
    else:
        assert response.status_code == 404


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('issue_id', issues)
@pytest.mark.parametrize('project_id', projects)
def test_issue_send_post_email(db, client, username, password, project_id, issue_id):
    client.login(username=username, password=password)
    issue = Issue.objects.filter(project_id=project_id, id=issue_id).first()

    url = reverse('issue_send', args=[project_id, issue_id])
    data = {
        'subject': 'Subject',
        'message': 'Message',
        'recipients': ['email@example.com']
    }
    response = client.post(url, data)

    if issue:
        if project_id in change_issue_permission_map.get(username, []):
            assert response.status_code == 302
            assert len(mail.outbox) == 1
            assert mail.outbox[0].subject == '[example.com] Subject'
            assert mail.outbox[0].body == 'Message'
        else:
            if password:
                assert response.status_code == 403
            else:
                assert response.status_code == 302

            assert len(mail.outbox) == 0
    else:
        assert response.status_code == 404


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('issue_id', issues)
@pytest.mark.parametrize('project_id', projects)
def test_issue_send_post_attachements(db, client, files, username, password, project_id, issue_id):
    client.login(username=username, password=password)
    issue = Issue.objects.filter(project_id=project_id, id=issue_id).first()

    if issue:
        view = issue.project.views.first()
        file = issue.project.values.filter(snapshot=None, value_type=VALUE_TYPE_FILE).first()

        if file and view:
            url = reverse('issue_send', args=[project_id, issue_id])
            data = {
                'subject': 'Subject',
                'message': 'Message',
                'recipients': 'email@example.com',
                'attachments_answers': 'project_answers',
                'attachments_views': str(view.id),
                'attachments_files': str(file.id),
                'attachments_snapshot': '',
                'attachments_format': 'html'
            }
            response = client.post(url, data)

            if project_id in change_issue_permission_map.get(username, []):
                assert response.status_code == 302
                assert len(mail.outbox) == 1
                assert mail.outbox[0].subject == '[example.com] Subject'
                assert mail.outbox[0].body == 'Message'

                attachments = mail.outbox[0].attachments
                assert len(attachments) == 3
                assert attachments[0][0] == 'Test.html'
                assert attachments[0][2] == 'text/html; charset=utf-8'
                assert attachments[1][0] == 'Test.html'
                assert attachments[1][2] == 'text/html; charset=utf-8'
                assert attachments[2][0] == 'test.txt'
                assert attachments[2][2] == 'text/plain'
            else:
                if password:
                    assert response.status_code == 403
                else:
                    assert response.status_code == 302

                assert len(mail.outbox) == 0


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('issue_id', issues)
@pytest.mark.parametrize('project_id', projects)
def test_issue_send_post_integration(db, client, username, password, project_id, issue_id):
    client.login(username=username, password=password)
    issue = Issue.objects.filter(project_id=project_id, id=issue_id).first()

    url = reverse('issue_send', args=[project_id, issue_id])
    data = {
        'subject': 'Subject',
        'message': 'Message',
        'integration': [integration_pk]
    }
    response = client.post(url, data)

    if issue:
        if project_id in change_issue_permission_map.get(username, []):
            if integration_pk in Project.objects.get(pk=project_id).integrations.values_list('id', flat=True):
                assert response.status_code == 302
                assert response.url.startswith('https://github.com')
            else:
                assert response.status_code == 200
        else:
            if password:
                assert response.status_code == 403
            else:
                assert response.status_code == 302
                assert not response.url.startswith('https://github.com')
    else:
        assert response.status_code == 404



import pytest
from django.core.exceptions import ValidationError
from rest_framework.exceptions import \
    ValidationError as RestFameworkValidationError

from ..models import OptionSet
from ..serializers.v1 import OptionSetSerializer
from ..validators import OptionSetLockedValidator


def test_create(db):
    OptionSetLockedValidator()({
        'locked': False
    })


def test_create_locked(db):
    OptionSetLockedValidator()({
        'locked': True
    })


def test_update(db):
    optionset = OptionSet.objects.first()

    OptionSetLockedValidator(optionset)({
        'locked': False
    })


def test_update_error(db):
    optionset = OptionSet.objects.first()
    optionset.locked = True
    optionset.save()

    with pytest.raises(ValidationError):
        OptionSetLockedValidator(optionset)({
            'locked': True
        })


def test_update_lock(db):
    optionset = OptionSet.objects.first()

    OptionSetLockedValidator(optionset)({
        'locked': True
    })


def test_update_unlock(db):
    optionset = OptionSet.objects.first()
    optionset.locked = True
    optionset.save()

    OptionSetLockedValidator(optionset)({
        'locked': False
    })


def test_serializer_create(db):
    validator = OptionSetLockedValidator()
    validator.set_context(OptionSetSerializer())

    validator({
        'locked': False
    })


def test_serializer_create_locked(db):
    validator = OptionSetLockedValidator()
    validator.set_context(OptionSetSerializer())

    validator({
        'locked': True
    })


def test_serializer_update(db):
    optionset = OptionSet.objects.first()

    validator = OptionSetLockedValidator()
    validator.set_context(OptionSetSerializer(instance=optionset))

    validator({})


def test_serializer_update_error(db):
    optionset = OptionSet.objects.first()
    optionset.locked = True
    optionset.save()

    validator = OptionSetLockedValidator()
    validator.set_context(OptionSetSerializer(instance=optionset))

    with pytest.raises(RestFameworkValidationError):
        validator({
            'locked': True
        })


def test_serializer_update_lock(db):
    optionset = OptionSet.objects.first()

    validator = OptionSetLockedValidator()
    validator.set_context(OptionSetSerializer(instance=optionset))

    validator({
        'locked': True
    })


def test_serializer_update_unlock(db):
    optionset = OptionSet.objects.first()
    optionset.locked = True
    optionset.save()

    validator = OptionSetLockedValidator()
    validator.set_context(OptionSetSerializer(instance=optionset))

    validator({
        'locked': False
    })



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-01-25 11:34
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('options', '0005_refactoring'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='optionset',
            options={'ordering': ('uri',), 'verbose_name': 'OptionSet', 'verbose_name_plural': 'OptionSets'},
        ),
        migrations.AddField(
            model_name='option',
            name='comment',
            field=models.TextField(blank=True, help_text='Additional information about this option.', null=True, verbose_name='Comment'),
        ),
        migrations.AddField(
            model_name='option',
            name='uri',
            field=models.URLField(blank=True, help_text='The Uniform Resource Identifier of this option (auto-generated).', max_length=640, null=True, verbose_name='URI'),
        ),
        migrations.AddField(
            model_name='option',
            name='uri_prefix',
            field=models.URLField(blank=True, help_text='The prefix for the URI of this option.', max_length=256, null=True, verbose_name='URI Prefix'),
        ),
        migrations.AddField(
            model_name='optionset',
            name='comment',
            field=models.TextField(blank=True, help_text='Additional information about this option set.', null=True, verbose_name='Comment'),
        ),
        migrations.AddField(
            model_name='optionset',
            name='uri',
            field=models.URLField(blank=True, help_text='The Uniform Resource Identifier of this option set (auto-generated).', max_length=640, null=True, verbose_name='URI'),
        ),
        migrations.AddField(
            model_name='optionset',
            name='uri_prefix',
            field=models.URLField(blank=True, help_text='The prefix for the URI of this option set.', max_length=256, null=True, verbose_name='URI Prefix'),
        ),
        migrations.AlterField(
            model_name='option',
            name='additional_input',
            field=models.BooleanField(default=False, help_text='Designates whether an additional input is possible for this option.', verbose_name='Additional input'),
        ),
        migrations.AlterField(
            model_name='option',
            name='key',
            field=models.SlugField(blank=True, help_text='The internal identifier of this option. The URI will be generated from this key.', max_length=128, null=True, verbose_name='Key'),
        ),
        migrations.AlterField(
            model_name='option',
            name='optionset',
            field=models.ForeignKey(blank=True, help_text='The option set this option belongs to.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='options', to='options.OptionSet', verbose_name='Option set'),
        ),
        migrations.AlterField(
            model_name='option',
            name='order',
            field=models.IntegerField(default=0, help_text='The position of this option in lists.', verbose_name='Order'),
        ),
        migrations.AlterField(
            model_name='option',
            name='text_de',
            field=models.CharField(help_text='The German text displayed for this option.', max_length=256, verbose_name='Text (de)'),
        ),
        migrations.AlterField(
            model_name='option',
            name='text_en',
            field=models.CharField(help_text='The English text displayed for this option.', max_length=256, verbose_name='Text (en)'),
        ),
        migrations.AlterField(
            model_name='optionset',
            name='conditions',
            field=models.ManyToManyField(blank=True, help_text='The list of conditions evaluated for this option set.', to='conditions.Condition', verbose_name='Conditions'),
        ),
        migrations.AlterField(
            model_name='optionset',
            name='key',
            field=models.SlugField(blank=True, help_text='The internal identifier of this option set. The URI will be generated from this key.', max_length=128, null=True, verbose_name='Key'),
        ),
        migrations.AlterField(
            model_name='optionset',
            name='order',
            field=models.IntegerField(default=0, help_text='The position of this option set in lists.', verbose_name='Order'),
        ),
    ]



import logging

from rest_framework import serializers
from rest_framework.reverse import reverse

from rdmo.conditions.models import Condition
from rdmo.questions.models import Question, QuestionSet
from rdmo.tasks.models import Task

from ..models import Attribute
from ..validators import (AttributeLockedValidator, AttributeParentValidator,
                          AttributeUniqueURIValidator)

log = logging.getLogger(__name__)


class ConditionSerializer(serializers.ModelSerializer):

    class Meta:
        model = Condition
        fields = (
            'id',
            'uri',
            'key',
        )


class QuestionSetSerializer(serializers.ModelSerializer):

    class Meta:
        model = QuestionSet
        fields = (
            'id',
            'uri',
            'path',
        )


class QuestionSerializer(serializers.ModelSerializer):

    class Meta:
        model = Question
        fields = (
            'id',
            'uri',
            'path',
        )


class TaskSerializer(serializers.ModelSerializer):

    class Meta:
        model = Task
        fields = (
            'id',
            'uri',
            'key',
        )


class AttributeSerializer(serializers.ModelSerializer):

    key = serializers.SlugField(required=True)
    parent = serializers.PrimaryKeyRelatedField(queryset=Attribute.objects.all(), default=None, allow_null=True)
    path = serializers.CharField(required=False)
    conditions = ConditionSerializer(many=True, read_only=True)
    questionsets = QuestionSetSerializer(many=True, read_only=True)
    questions = QuestionSerializer(many=True, read_only=True)
    tasks_as_start = TaskSerializer(many=True, read_only=True)
    tasks_as_end = TaskSerializer(many=True, read_only=True)
    values_count = serializers.IntegerField(read_only=True)
    projects_count = serializers.IntegerField(read_only=True)

    class Meta:
        model = Attribute
        fields = (
            'id',
            'uri',
            'uri_prefix',
            'key',
            'path',
            'comment',
            'locked',
            'parent',
            'conditions',
            'questionsets',
            'questions',
            'tasks_as_start',
            'tasks_as_end',
            'values_count',
            'projects_count'
        )
        validators = (
            AttributeUniqueURIValidator(),
            AttributeParentValidator(),
            AttributeLockedValidator()
        )


class AttributeNestedSerializer(serializers.ModelSerializer):

    children = serializers.SerializerMethodField()
    xml_url = serializers.SerializerMethodField()

    class Meta:
        model = Attribute
        fields = (
            'id',
            'uri',
            'uri_prefix',
            'path',
            'key',
            'locked',
            'children',
            'xml_url'
        )

    def get_children(self, obj):
        # get the children from the cached mptt tree
        return AttributeNestedSerializer(obj.get_children(), many=True, read_only=True).data

    def get_xml_url(self, obj):
        return reverse('v1-domain:attribute-detail-export', args=[obj.pk])


class AttributeIndexSerializer(serializers.ModelSerializer):

    class Meta:
        model = Attribute
        fields = (
            'id',
            'uri',
            'key',
            'path'
        )



from django.utils.translation import gettext_lazy as _

from rdmo.core.validators import LockedValidator, UniqueURIValidator

from .models import Option, OptionSet


class OptionSetUniqueURIValidator(UniqueURIValidator):

    model = OptionSet

    def get_uri(self, data):
        if not data.get('key'):
            self.raise_validation_error({'key': _('This field is required.')})
        else:
            uri = self.model.build_uri(data.get('uri_prefix'), data.get('key'))
            return uri


class OptionUniqueURIValidator(UniqueURIValidator):

    model = Option

    def get_uri(self, data):
        if not data.get('key'):
            self.raise_validation_error({'key': _('This field is required.')})
        elif not data.get('optionset'):
            self.raise_validation_error({'optionset': _('This field may not be null.')})
        else:
            path = self.model.build_path(data.get('key'), data.get('optionset'))
            uri = self.model.build_uri(data.get('uri_prefix'), path)
            return uri


class OptionSetLockedValidator(LockedValidator):

    pass


class OptionLockedValidator(LockedValidator):

    parent_field = 'optionset'



import os
import re
from pathlib import Path

import pytest
from django.urls import reverse

from rdmo.core.constants import VALUE_TYPE_FILE

from ..models import Project, Value

users = (
    ('owner', 'owner'),
    ('manager', 'manager'),
    ('author', 'author'),
    ('guest', 'guest'),
    ('user', 'user'),
    ('site', 'site'),
    ('anonymous', None),
)

view_project_permission_map = {
    'owner': [1, 2, 3, 4, 5],
    'manager': [1, 3, 5],
    'author': [1, 3, 5],
    'guest': [1, 3, 5],
    'api': [1, 2, 3, 4, 5],
    'site': [1, 2, 3, 4, 5]
}

change_project_permission_map = {
    'owner': [1, 2, 3, 4, 5],
    'manager': [1, 3, 5],
    'api': [1, 2, 3, 4, 5],
    'site': [1, 2, 3, 4, 5]
}

projects = [1, 2, 3, 4, 5]


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
def test_project_update_import_get(db, client, username, password, project_id):
    client.login(username=username, password=password)

    url = reverse('project_update_import', args=[project_id])
    response = client.get(url)

    if project_id in change_project_permission_map.get(username, []):
        assert response.status_code == 302
        assert response.url == '/projects/{}/'.format(project_id)
    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
def test_project_update_import_post_error(db, settings, client, username, password, project_id):
    client.login(username=username, password=password)

    url = reverse('project_update_import', args=[project_id])
    response = client.post(url, {
        'method': 'wrong'
    })

    if project_id in change_project_permission_map.get(username, []):
        assert response.status_code == 400
    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
def test_project_update_import_post_upload_file(db, settings, client, username, password, project_id):
    client.login(username=username, password=password)

    url = reverse('project_update_import', args=[project_id])
    xml_file = os.path.join(settings.BASE_DIR, 'xml', 'project.xml')
    with open(xml_file, encoding='utf8') as f:
        response = client.post(url, {
            'method': 'upload_file',
            'uploaded_file': f
        })

        if project_id in change_project_permission_map.get(username, []):
            assert response.status_code == 200
            assert b'Import from project.xml' in response.content
        elif password:
            assert response.status_code == 403
        else:
            assert response.status_code == 302
            assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
def test_project_update_import_post_upload_file_error(db, settings, client, username, password, project_id):
    client.login(username=username, password=password)

    url = reverse('project_update_import', args=[project_id])
    xml_file = os.path.join(settings.BASE_DIR, 'xml', 'error.xml')
    with open(xml_file, encoding='utf8') as f:
        response = client.post(url, {
            'method': 'upload_file',
            'uploaded_file': f
        })

    if project_id in change_project_permission_map.get(username, []):
        assert response.status_code == 400
        assert b'Files of this type cannot be imported.' in response.content
    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
def test_project_update_import_post_upload_file_empty(db, client, username, password, project_id):
    client.login(username=username, password=password)

    url = reverse('project_update_import', args=[project_id])
    response = client.post(url, {
        'method': 'upload_file'
    })

    if project_id in change_project_permission_map.get(username, []):
        assert response.status_code == 400
        assert b'There has been an error with your import.' in response.content
    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
def test_project_update_import_post_import_file(db, settings, client, files, username, password, project_id):
    client.login(username=username, password=password)
    projects_count = Project.objects.count()

    project = Project.objects.get(pk=project_id)
    project_updated = project.updated
    project_snapshot_count = project.snapshots.count()
    project_snapshot_values_count = project.values.filter(snapshot=None).count()
    project_values_count = project.values.count()

    # upload file
    url = reverse('project_update_import', args=[project_id])
    xml_file = os.path.join(settings.BASE_DIR, 'xml', 'project.xml')
    with open(xml_file, encoding='utf8') as f:
        response = client.post(url,  {
            'method': 'upload_file',
            'uploaded_file': f
        })

    if project_id in change_project_permission_map.get(username, []):
        assert response.status_code == 200

        # get keys from the response
        keys = re.findall(r'name=\"(.*?)\"', response.content.decode())

        # import file
        url = reverse('project_update_import', args=[project_id])
        data = {key: ['on'] for key in keys}
        data.update({'method': 'import_file'})
        response = client.post(url, data)

            # check if all the files are where are supposed to be
        for file_value in Value.objects.filter(value_type=VALUE_TYPE_FILE):
            assert Path(settings.MEDIA_ROOT).joinpath(file_value.file.name).exists()

        # no new project, snapshots, values were created
        project = Project.objects.get(pk=project_id)

        assert Project.objects.count() == projects_count
        assert project.snapshots.count() == project_snapshot_count
        if project_id == 1:
            assert project.values.count() == project_values_count
            assert project.values.filter(snapshot=None).count() == project_snapshot_values_count

        assert project.updated == project_updated

        if project_id in change_project_permission_map.get(username, []):
            assert response.status_code == 302
            assert response.url == '/projects/{}/'.format(project_id)

        else:
            if password:
                assert response.status_code == 403
            else:
                assert response.status_code == 302
                assert response.url.startswith('/account/login/')

    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
def test_project_update_import_post_import_file_empty(db, settings, client, username, password, project_id):
    client.login(username=username, password=password)
    projects_count = Project.objects.count()

    project = Project.objects.get(pk=project_id)
    project_updated = project.updated
    project_snapshot_count = project.snapshots.count()
    project_snapshot_values_count = project.values.filter(snapshot=None).count()
    project_values_count = project.values.count()

    # upload file
    url = reverse('project_update_import', args=[project_id])
    xml_file = os.path.join(settings.BASE_DIR, 'xml', 'project.xml')
    with open(xml_file, encoding='utf8') as f:
        response = client.post(url, {
            'method': 'upload_file',
            'uploaded_file': f
        })

    if project_id in change_project_permission_map.get(username, []):
        assert response.status_code == 200, project_id

        response = client.post(url, {
            'method': 'import_file'
        })

            # check if all the files are where are supposed to be
        for file_value in Value.objects.filter(value_type=VALUE_TYPE_FILE):
            assert Path(settings.MEDIA_ROOT).joinpath(file_value.file.name).exists()

        # no new project, snapshots, values were created
        project = Project.objects.get(pk=project_id)
        assert Project.objects.count() == projects_count
        assert project.snapshots.count() == project_snapshot_count
        assert project.values.count() == project_values_count
        assert project.values.filter(snapshot=None).count() == project_snapshot_values_count
        assert project.updated == project_updated

        assert response.status_code == 302
        assert response.url == '/projects/{}/'.format(project_id)
    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
@pytest.mark.parametrize('source_id', projects)
def test_project_update_import_post_import_project_step1(db, settings, client, username, password, project_id, source_id):
    client.login(username=username, password=password)

    url = reverse('project_update_import', args=[project_id])
    response = client.post(url, {
        'method': 'import_project',
        'source': source_id
    })

    if project_id in change_project_permission_map.get(username, []):
        if source_id in view_project_permission_map.get(username, []):
            assert response.status_code == 200
        else:
            assert response.status_code == 403

    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
@pytest.mark.parametrize('source_id', projects)
def test_project_update_import_post_import_project_step2(db, settings, client, username, password, project_id, source_id):
    client.login(username=username, password=password)
    projects_count = Project.objects.count()

    project = Project.objects.get(pk=project_id)
    project_updated = project.updated
    project_snapshot_count = project.snapshots.count()
    project_snapshot_values_count = project.values.filter(snapshot=None).count()
    project_values_count = project.values.count()

    source = Project.objects.get(pk=source_id)
    source_snapshot_count = source.snapshots.count()
    source_snapshot_values_count = source.values.filter(snapshot=None).count()
    source_values_count = source.values.count()

    url = reverse('project_update_import', args=[project_id])
    response = client.post(url, {
        'method': 'import_project',
        'source': source_id
    })

    if project_id in change_project_permission_map.get(username, []):
        if source_id in view_project_permission_map.get(username, []):
            assert response.status_code == 200

            # get keys from the response
            keys = re.findall(r'name=\"(.*?)\"', response.content.decode())

            # import file
            url = reverse('project_update_import', args=[project_id])
            data = {key: ['on'] for key in keys}
            data.update({
                'method': 'import_project',
                'source': source_id
            })
            response = client.post(url, data)

            # check if all the files are where are supposed to be
            for file_value in Value.objects.filter(value_type=VALUE_TYPE_FILE):
                assert Path(settings.MEDIA_ROOT).joinpath(file_value.file.name).exists()

            # no new project, snapshots, values were created
            project = Project.objects.get(pk=project_id)
            source = Project.objects.get(pk=source_id)

            # no new project was created
            assert Project.objects.count() == projects_count

            # the project has the correct count of snapshot and values
            assert project.snapshots.count() == project_snapshot_count
            if project_id == 1:
                assert project.values.count() == project_values_count
                assert project.values.filter(snapshot=None).count() == project_snapshot_values_count

            # the source project has the correct count of snapshot and values
            assert source.snapshots.count() == source_snapshot_count
            if source_id == 1:
                assert source.values.count() == source_values_count
                assert source.values.filter(snapshot=None).count() == source_snapshot_values_count

            assert project.updated == project_updated

            if project_id in change_project_permission_map.get(username, []):
                assert response.status_code == 302
                assert response.url == '/projects/{}/'.format(project_id)

            else:
                if password:
                    assert response.status_code == 403
                else:
                    assert response.status_code == 302
                    assert response.url.startswith('/account/login/')
        else:
            assert response.status_code == 403

    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')


@pytest.mark.parametrize('username,password', users)
@pytest.mark.parametrize('project_id', projects)
def test_project_update_import_post_import_project_error(db, settings, client, username, password, project_id):
    client.login(username=username, password=password)

    # upload file
    url = reverse('project_update_import', args=[project_id])
    response = client.post(url, {
        'method': 'import_project'
    })

    if project_id in change_project_permission_map.get(username, []):
        assert response.status_code == 404
    elif password:
        assert response.status_code == 403
    else:
        assert response.status_code == 302
        assert response.url.startswith('/account/login/')



import pytest
from django.template import Context

from rdmo.projects.models import Project
from rdmo.views.templatetags.view_tags import (get_set_value, get_set_values,
                                               get_sets, get_value, get_values)
from rdmo.views.utils import ProjectWrapper

project_pk = 1


def assertListEqual(value_list, queryset):
    assert [value['id'] for value in value_list] == list(queryset.values_list('id', flat=True))


@pytest.fixture
def context(db):
    return Context({
        'project': ProjectWrapper(Project.objects.get(pk=project_pk))
    })


@pytest.fixture
def values(db):
    return Project.objects.get(pk=project_pk).values.filter(snapshot=None).order_by('set_index').order_by('collection_index')


def test_get_value_project_title(context):
    project = Project.objects.get(pk=project_pk)
    assert get_value(context, 'project/title')['value'] == project.title


def test_get_value_project_description(context):
    project = Project.objects.get(pk=project_pk)
    assert get_value(context, 'project/description')['value'] == project.description


def test_get_value_project_created(context):
    project = Project.objects.get(pk=project_pk)
    assert get_value(context, 'project/created')['value'] == project.created


def test_get_value_project_updated(context):
    project = Project.objects.get(pk=project_pk)
    assert get_value(context, 'project/updated')['value'] == project.updated


def test_get_values_uri(context, values):
    uri = 'http://example.com/terms/domain/individual/collection/text'
    assertListEqual(get_values(context, uri), values.filter(attribute__uri=uri))


def test_get_values(context, values):
    path = 'individual/collection/text'
    assertListEqual(get_values(context, path), values.filter(attribute__path=path))


def test_get_values_empty(context):
    assert get_values(context, 'individual') == []


def test_get_values_wrong(context):
    assert get_values(context, 'wrong') == []


def test_get_values_set_index(context, values):
    path = 'set/single/text'
    assertListEqual(get_values(context, path, set_index=1), values.filter(attribute__path=path, set_index=1))


def test_get_values_collection_index(context, values):
    path = 'individual/collection/text'
    assertListEqual(get_values(context, path, index=1), values.filter(attribute__path=path, collection_index=1))


def test_get_value(context, values):
    path = 'individual/collection/text'
    assert get_value(context, path)['id'] == values.filter(attribute__path=path).first().id


def test_get_set_values(context, values):
    path = 'set/single/text'
    for value_set in get_sets(context, 'set'):
        assertListEqual(get_set_values(context, value_set, path), values.filter(attribute__path=path, set_index=value_set['set_index']))


def test_get_set_value(context, values):
    path = 'set/single/text'
    for value_set in get_sets(context, 'set'):
        assert get_set_value(context, value_set, path)['id'] == values.filter(attribute__path=path, set_index=value_set['set_index']).first().id



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-03-13 11:22
from __future__ import unicode_literals

from django.db import migrations


def run_data_migration(apps, schema_editor):
    Option = apps.get_model('options', 'Option')

    for option in Option.objects.all():
        if option.optionset == None:
            option.delete()


class Migration(migrations.Migration):

    dependencies = [
        ('options', '0020_django2'),
    ]

    operations = [
        migrations.RunPython(run_data_migration),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-02-03 11:30
from __future__ import unicode_literals

from django.db import migrations, models
from django.utils.text import slugify


def migrate_views(apps, schema_editor):

    View = apps.get_model('views', 'view')

    for view in View.objects.all():
        view.title_en = view.key
        view.title_de = view.key
        view.key = slugify(view.key)
        view.save()


class Migration(migrations.Migration):

    dependencies = [
        ('views', '0006_title_and_help'),
    ]

    operations = [
        migrations.RunPython(migrate_views),
    ]



from django import template
from django.template.defaultfilters import stringfilter
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from rdmo.views.templatetags.view_tags import get_value

from ..models import Membership

register = template.Library()


@register.simple_tag()
def projects_indent(level):
    string = ''
    if level > 0:
        for _ in range(level - 1):
            string += '&ensp;&ensp;'
        string += '&#8226;&ensp;'

    return mark_safe('<span class="projects-indent">' + string + '</span>')


@register.filter()
@stringfilter
def projects_role(role):
    return dict(Membership.ROLE_CHOICES).get(role, '')


@register.simple_tag(takes_context=True)
def get_labels(context, question, set_prefix='', set_index=0, project=None):
    if question.questionset.is_collection:
        set_labels = []

        for questionset in question.questionset.get_ancestors(ascending=True, include_self=True):
            set_label = '#{}'.format(set_index + 1)

            if questionset.attribute:
                # get attribute value
                value = get_value(context, questionset.attribute.uri, set_prefix=set_prefix, set_index=set_index, index=0, project=project)
                if value:
                    set_label = '"{}"'.format(value['value'])

            set_labels.append('{} {}'.format(questionset.verbose_name.title() or _('Set'), set_label))

            if set_prefix != '':
                rpartition = set_prefix.rpartition('|')
                set_prefix, set_index = rpartition[0], int(rpartition[2])

        # flip the list
        set_labels.reverse()

        return set_labels
    else:
        return None


@register.simple_tag(takes_context=True)
def check_question(context, question, set_prefix=None, set_index=None, project=None):
    if project is None:
        project = context['project']

    conditions = list(filter(lambda condition: question in condition.questions.all(), project._conditions))
    return project._check_conditions(conditions, set_prefix=set_prefix, set_index=set_index)


@register.simple_tag(takes_context=True)
def check_questionset(context, questionset, set_prefix=None, set_index=None, project=None):
    if project is None:
        project = context['project']

    conditions = list(filter(lambda condition: questionset in condition.questionsets.all(), project._conditions))
    return project._check_conditions(conditions, set_prefix=set_prefix, set_index=set_index)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def run_data_migration(apps, schema_editor):
    Task = apps.get_model('tasks', 'Task')
    Timeframe = apps.get_model('tasks', 'Timeframe')

    # migrate questionsets
    for task in Task.objects.all():
        try:
            task.start_attribute = task.timeframe.start_attribute
            task.end_attribute = task.timeframe.end_attribute
            task.days_before = task.timeframe.days_before
            task.days_after = task.timeframe.days_after
            task.save()

        except Timeframe.DoesNotExist:
            pass


class Migration(migrations.Migration):

    dependencies = [
        ('tasks', '0014_move_timeframe_to_task'),
    ]

    operations = [
        migrations.RunPython(run_data_migration),
    ]



import logging

from django.contrib.sites.models import Site
from django.http import (HttpResponseBadRequest, HttpResponseForbidden,
                         HttpResponseRedirect)
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.urls import reverse
from django.views.generic import DeleteView, UpdateView
from django.views.generic.edit import FormView

from rdmo.accounts.utils import is_site_manager
from rdmo.core.mail import send_mail
from rdmo.core.views import ObjectPermissionMixin, RedirectViewMixin

from ..forms import MembershipCreateForm
from ..models import Membership, Project
from ..utils import is_last_owner

logger = logging.getLogger(__name__)


class MembershipCreateView(ObjectPermissionMixin, RedirectViewMixin, FormView):
    model = Membership
    form_class = MembershipCreateForm
    permission_required = 'projects.add_membership_object'
    template_name = 'projects/membership_form.html'

    def dispatch(self, *args, **kwargs):
        self.project = get_object_or_404(Project.objects.all(), pk=self.kwargs['project_id'])
        return super().dispatch(*args, **kwargs)

    def get_queryset(self):
        return Membership.objects.filter(project=self.project)

    def get_permission_object(self):
        return self.project

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['project'] = self.project
        kwargs['is_site_manager'] = is_site_manager(self.request.user)
        return kwargs

    def get_success_url(self):
        return self.project.get_absolute_url()

    def form_valid(self, form):
        invite = form.save()
        if invite is not None:
            context = {
                'invite_url': self.request.build_absolute_uri(reverse('project_join', args=[invite.token])),
                'invite_user': invite.user,
                'project': invite.project,
                'user': self.request.user,
                'site': Site.objects.get_current()
            }

            subject = render_to_string('projects/email/project_invite_subject.txt', context)
            message = render_to_string('projects/email/project_invite_message.txt', context)

            # send the email
            send_mail(subject, message, to=[invite.email])

        return super().form_valid(form)


class MembershipUpdateView(ObjectPermissionMixin, RedirectViewMixin, UpdateView):
    fields = ('role', )
    permission_required = 'projects.change_membership_object'

    def get_queryset(self):
        return Membership.objects.filter(project_id=self.kwargs.get('project_id'))

    def get_permission_object(self):
        return self.get_object().project


class MembershipDeleteView(ObjectPermissionMixin, RedirectViewMixin, DeleteView):
    permission_required = 'projects.delete_membership_object'

    def get_queryset(self):
        return Membership.objects.filter(project_id=self.kwargs.get('project_id'))

    def delete(self, *args, **kwargs):
        self.obj = self.get_object()

        if (self.request.user in self.obj.project.owners) or is_site_manager(self.request.user):
            # user is owner or site manager
            if is_last_owner(self.obj.project, self.obj.user):
                logger.info('User "%s" not allowed to remove last user "%s"', self.request.user.username, self.obj.user.username)
                return HttpResponseBadRequest()
            else:
                logger.info('User "%s" deletes user "%s"', self.request.user.username, self.obj.user.username)
                success_url = reverse('project', args=[self.get_object().project.id])
                self.obj.delete()
                return HttpResponseRedirect(success_url)

        elif self.request.user == self.obj.user:
            # user wants to remove him/herself
            logger.info('User "%s" deletes himself.', self.request.user.username)
            success_url = reverse('projects')
            self.obj.delete()
            return HttpResponseRedirect(success_url)

        else:
            logger.info('User "%s" not allowed to remove user "%s"', self.request.user.username, self.obj.user.username)
            return HttpResponseForbidden()

    def get_permission_object(self):
        return self.get_object().project



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('questions', '0001_initial_after_reset'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='question',
            options={'verbose_name': 'Question', 'verbose_name_plural': 'Questions'},
        ),
        migrations.AlterModelOptions(
            name='questionentity',
            options={'ordering': ('subsection__section__catalog__order', 'subsection__section__order', 'subsection__order', 'order'), 'verbose_name': 'QuestionEntity', 'verbose_name_plural': 'QuestionEntities'},
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-03-13 11:54
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('projects', '0024_data_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='value',
            name='text',
            field=models.TextField(blank=True, help_text='The string stored for this value.', verbose_name='Text'),
        ),
    ]



import pytest
from django.core.exceptions import ValidationError
from rest_framework.exceptions import \
    ValidationError as RestFameworkValidationError

from ..models import View
from ..serializers.v1 import ViewSerializer
from ..validators import ViewLockedValidator


def test_create(db):
    ViewLockedValidator()({
        'locked': False
    })


def test_create_locked(db):
    ViewLockedValidator()({
        'locked': True
    })


def test_update(db):
    view = View.objects.first()

    ViewLockedValidator(view)({
        'locked': False
    })


def test_update_error(db):
    view = View.objects.first()
    view.locked = True
    view.save()

    with pytest.raises(ValidationError):
        ViewLockedValidator(view)({
            'locked': True
        })


def test_update_lock(db):
    view = View.objects.first()

    ViewLockedValidator(view)({
        'locked': True
    })


def test_update_unlock(db):
    view = View.objects.first()
    view.locked = True
    view.save()

    ViewLockedValidator(view)({
        'locked': False
    })


def test_serializer_create(db):
    validator = ViewLockedValidator()
    validator.set_context(ViewSerializer())

    validator({
        'locked': False
    })


def test_serializer_create_locked(db):
    validator = ViewLockedValidator()
    validator.set_context(ViewSerializer())

    validator({
        'locked': True
    })


def test_serializer_update(db):
    view = View.objects.first()

    validator = ViewLockedValidator()
    validator.set_context(ViewSerializer(instance=view))

    validator({})


def test_serializer_update_error(db):
    view = View.objects.first()
    view.locked = True
    view.save()

    validator = ViewLockedValidator()
    validator.set_context(ViewSerializer(instance=view))

    with pytest.raises(RestFameworkValidationError):
        validator({
            'locked': True
        })


def test_serializer_update_lock(db):
    view = View.objects.first()

    validator = ViewLockedValidator()
    validator.set_context(ViewSerializer(instance=view))

    validator({
        'locked': True
    })


def test_serializer_update_unlock(db):
    view = View.objects.first()
    view.locked = True
    view.save()

    validator = ViewLockedValidator()
    validator.set_context(ViewSerializer(instance=view))

    validator({
        'locked': False
    })



from django.urls import reverse


def test_optionset_search(db, client):
    client.login(username='admin', password='admin')

    url = reverse('admin:options_optionset_changelist') + '?q=test'
    response = client.get(url)
    assert response.status_code == 200


def test_option_search(db, client):
    client.login(username='admin', password='admin')

    url = reverse('admin:options_option_changelist') + '?q=test'
    response = client.get(url)
    assert response.status_code == 200



from django.contrib.sites.models import Site
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet

from .models import Overlay
from .utils import get_overlays


class OverlayViewSet(ViewSet):

    @action(detail=False, methods=['post'], url_path='(?P<url_name>[-\\w]+)/current', permission_classes=[IsAuthenticated])
    def current(self, request, url_name=None):
        site = Site.objects.get_current()
        overlays = get_overlays(url_name)
        if not overlays:
            raise NotFound()

        try:
            overlay = Overlay.objects.get(user=request.user, site=site, url_name=url_name)
        except Overlay.DoesNotExist:
            overlay = Overlay.objects.create(user=request.user, site=site, url_name=url_name, current=overlays[0])

        return Response({
            'overlay': overlay.current,
            'last': overlay.current == overlays[-1]
        })

    @action(detail=False, methods=['post'], url_path='(?P<url_name>[-\\w]+)/next', permission_classes=[IsAuthenticated])
    def next(self, request, url_name=None):
        site = Site.objects.get_current()
        overlays = get_overlays(url_name)
        if not overlays:
            raise NotFound()

        try:
            overlay = Overlay.objects.get(user=request.user, site=site, url_name=url_name)
            try:
                overlay.current = overlays[overlays.index(overlay.current) + 1]
            except (IndexError, ValueError):
                overlay.current = ''
            overlay.save()
        except Overlay.DoesNotExist:
            overlay = Overlay.objects.create(user=request.user, site=site, url_name=url_name, current=overlays[0])

        return Response({
            'overlay': overlay.current,
            'last': overlay.current == overlays[-1]
        })

    @action(detail=False, methods=['post'], url_path='(?P<url_name>[-\\w]+)/dismiss', permission_classes=[IsAuthenticated])
    def dismiss(self, request, url_name=None):
        site = Site.objects.get_current()
        overlays = get_overlays(url_name)
        if not overlays:
            raise NotFound()

        try:
            overlay = Overlay.objects.get(user=request.user, site=site, url_name=url_name)
            overlay.current = ''
            overlay.save()
        except Overlay.DoesNotExist:
            overlay = Overlay.objects.create(user=request.user, site=site, url_name=url_name, current='')

        return Response({'overlay': overlay.current})



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-08-03 09:06
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.db.models.manager
import mptt.fields


class Migration(migrations.Migration):

    dependencies = [
        ('domain', '0012_renaming'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='attribute',
            managers=[
                ('objects', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name='attributeentity',
            managers=[
                ('objects', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AddField(
            model_name='attributeentity',
            name='level',
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='attributeentity',
            name='lft',
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='attributeentity',
            name='rght',
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='attributeentity',
            name='tree_id',
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='attributeentity',
            name='parent',
            field=mptt.fields.TreeForeignKey(blank=True, help_text='optional', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='domain.AttributeEntity'),
        ),
    ]



import logging

from django.conf import settings
from django.db.models import Prefetch
from django.views.generic import DetailView, TemplateView

from rdmo.core.exports import XMLResponse
from rdmo.core.utils import get_model_field_meta, render_to_format
from rdmo.core.views import CSRFViewMixin, ModelPermissionMixin

from .models import Catalog, Question, QuestionSet, Section
from .renderers import CatalogRenderer
from .serializers.export import CatalogExportSerializer

log = logging.getLogger(__name__)


class CatalogsView(ModelPermissionMixin, CSRFViewMixin, TemplateView):
    template_name = 'questions/catalogs.html'
    permission_required = 'questions.view_catalog'

    def get_context_data(self, **kwargs):
        context = super(CatalogsView, self).get_context_data(**kwargs)
        context['export_formats'] = settings.EXPORT_FORMATS
        context['meta'] = {
            'Catalog': get_model_field_meta(Catalog),
            'Section': get_model_field_meta(Section),
            'QuestionSet': get_model_field_meta(QuestionSet),
            'Question': get_model_field_meta(Question),
        }
        return context


class CatalogExportView(ModelPermissionMixin, DetailView):
    model = Catalog
    context_object_name = 'catalog'
    permission_required = 'questions.view_catalog'

    def get_queryset(self):
        return Catalog.objects.prefetch_related(
            'sections',
            Prefetch('sections__questionsets', queryset=QuestionSet.objects.filter(questionset=None).prefetch_related(
                'conditions',
                'questions',
                'questions__attribute',
                'questions__optionsets',
                'questionsets',
                'questionsets__attribute',
                'questionsets__conditions',
                'questionsets__questions',
                'questionsets__questions__attribute',
                'questionsets__questions__optionsets',
                'questionsets__questionsets'
            ).select_related('attribute'))
        )

    def render_to_response(self, context, **response_kwargs):
        format = self.kwargs.get('format')
        if format == 'xml':
            serializer = CatalogExportSerializer(context['catalog'])
            xml = CatalogRenderer().render([serializer.data])
            return XMLResponse(xml, name=context['catalog'].key)
        else:
            return render_to_format(self.request, format, context['catalog'].title, 'questions/catalog_export.html', context)



from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations


def run_data_migration(apps, schema_editor):
    Task = apps.get_model('tasks', 'Task')
    Site = apps.get_model('sites', 'Site')

    for task in Task.objects.all():
        if not task.sites.exists():
            task.sites.add(Site.objects.get(pk=settings.SITE_ID))


class Migration(migrations.Migration):

    dependencies = [
        ('tasks', '0027_manager'),
    ]

    operations = [
        migrations.RunPython(run_data_migration),
    ]



from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from ..managers import MembershipManager


class Membership(models.Model):

    objects = MembershipManager()

    ROLE_CHOICES = (
        ('owner', _('Owner')),
        ('manager', _('Manager')),
        ('author', _('Author')),
        ('guest', _('Guest')),
    )

    project = models.ForeignKey(
        'Project', on_delete=models.CASCADE, related_name='memberships',
        verbose_name=_('Project'),
        help_text=_('The project for this membership.')
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='memberships',
        verbose_name=_('User'),
        help_text=_('The user for this membership.')
    )
    role = models.CharField(
        max_length=12, choices=ROLE_CHOICES,
        verbose_name=_('Role'),
        help_text=_('The role for this membership.')
    )

    class Meta:
        ordering = ('project__title', )
        verbose_name = _('Membership')
        verbose_name_plural = _('Memberships')

    def __str__(self):
        return '%s / %s / %s' % (self.project.title, self.user.username, self.role)

    def get_absolute_url(self):
        return reverse('project', kwargs={'pk': self.project.pk})

    @property
    def is_last_owner(self):
        return not Membership.objects.filter(project=self.project, role='owner').exclude(user=self.user).exists()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DetailKey',
            fields=[
                ('id', models.AutoField(verbose_name='ID', auto_created=True, serialize=False, primary_key=True)),
                ('key', models.SlugField()),
                ('label', models.CharField(max_length=256)),
                ('type', models.CharField(choices=[('text', 'Input field'), ('textarea', 'Textarea field'), ('checkbox', 'Checkbox'), ('radio', 'Radio button'), ('select', 'Select field'), ('multiselect', 'Multiselect field')], max_length=8)),
                ('hint', models.TextField(blank=True)),
                ('options', jsonfield.fields.JSONField(null=True, blank=True)),
                ('required', models.BooleanField()),
            ],
        ),
        migrations.CreateModel(
            name='Profile',
            fields=[
                ('id', models.AutoField(verbose_name='ID', auto_created=True, serialize=False, primary_key=True)),
                ('details', jsonfield.fields.JSONField(null=True, blank=True)),
                ('user', models.OneToOneField(to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('user',),
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-11-14 12:13
from __future__ import unicode_literals

try:
    import rdmo.core.models
except ImportError:
    import rdmo.core.models

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('accounts', '0006_permissions_removed'),
    ]

    operations = [
        migrations.CreateModel(
            name='AdditionalField',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.SlugField()),
                ('type', models.CharField(choices=[('text', 'Text'), ('textarea', 'Textarea')], max_length=11)),
                ('text_en', models.CharField(max_length=256)),
                ('text_de', models.CharField(max_length=256)),
                ('help_en', models.TextField(blank=True, help_text='Enter a help text to be displayed next to the input element', null=True)),
                ('help_de', models.TextField(blank=True, help_text='Enter a help text to be displayed next to the input element', null=True)),
                ('required', models.BooleanField()),
            ],
            options={
                'ordering': ('key',),
                'verbose_name': 'Additional field',
                'verbose_name_plural': 'Additional fields',
            },
            bases=(models.Model, rdmo.core.models.TranslationMixin),
        ),
        migrations.CreateModel(
            name='AdditionalFieldValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=256)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='accounts.AdditionalField')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='additional_fields', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('user', 'field'),
                'verbose_name': 'Additional field value',
                'verbose_name_plural': 'Additional field values',
            },
        ),
        migrations.DeleteModel(
            name='DetailKey',
        ),
        migrations.RemoveField(
            model_name='profile',
            name='user',
        ),
        migrations.DeleteModel(
            name='Profile',
        ),
    ]



# Generated by Django 2.2.13 on 2020-09-25 09:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('projects', '0039_integrationoption_secret'),
    ]

    operations = [
        migrations.CreateModel(
            name='IssueResource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.URLField(help_text='The URL of this issue resource.', verbose_name='URL')),
                ('integration', models.ForeignKey(help_text='The integration for this issue resource.', on_delete=django.db.models.deletion.CASCADE, related_name='resources', to='projects.Integration', verbose_name='Integration')),
                ('issue', models.ForeignKey(help_text='The issue for this issue resource.', on_delete=django.db.models.deletion.CASCADE, related_name='resources', to='projects.Issue', verbose_name='Issue')),
            ],
            options={
                'verbose_name': 'Issue resource',
                'verbose_name_plural': 'Issue resources',
                'ordering': ('issue__project__title', 'issue'),
            },
        ),
    ]



from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from rdmo.conditions.models import Condition
from rdmo.core.constants import VALUE_TYPE_CHOICES
from rdmo.core.models import Model, TranslationMixin
from rdmo.core.utils import copy_model, join_url
from rdmo.domain.models import Attribute
from rdmo.options.models import Option

from ..managers import QuestionManager
from .questionset import QuestionSet


class Question(Model, TranslationMixin):

    objects = QuestionManager()

    uri = models.URLField(
        max_length=640, blank=True, null=True,
        verbose_name=_('URI'),
        help_text=_('The Uniform Resource Identifier of this question (auto-generated).')
    )
    uri_prefix = models.URLField(
        max_length=256,
        verbose_name=_('URI Prefix'),
        help_text=_('The prefix for the URI of this question.')
    )
    key = models.SlugField(
        max_length=128, blank=True, null=True,
        verbose_name=_('Key'),
        help_text=_('The internal identifier of this question.')
    )
    path = models.CharField(
        max_length=512, blank=True, null=True,
        verbose_name=_('Path'),
        help_text=_('The path part of the URI of this question (auto-generated).')
    )
    comment = models.TextField(
        blank=True, null=True,
        verbose_name=_('Comment'),
        help_text=_('Additional internal information about this question.')
    )
    locked = models.BooleanField(
        default=False,
        verbose_name=_('Locked'),
        help_text=_('Designates whether this question can be changed.')
    )
    attribute = models.ForeignKey(
        Attribute, blank=True, null=True, on_delete=models.SET_NULL, related_name='questions',
        verbose_name=_('Attribute'),
        help_text=_('The attribute this question belongs to.')
    )
    questionset = models.ForeignKey(
        QuestionSet, on_delete=models.CASCADE, related_name='questions',
        verbose_name=_('Questionset'),
        help_text=_('The question set this question belongs to.')
    )
    is_collection = models.BooleanField(
        default=False,
        verbose_name=_('is collection'),
        help_text=_('Designates whether this question is a collection.')
    )
    is_optional = models.BooleanField(
        default=False,
        verbose_name=_('is optional'),
        help_text=_('Designates whether this question is optional.')
    )
    order = models.IntegerField(
        default=0,
        verbose_name=_('Order'),
        help_text=_('The position of this question in lists.')
    )
    help_lang1 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Help (primary)'),
        help_text=_('The help text for this question in the primary language.')
    )
    help_lang2 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Help (secondary)'),
        help_text=_('The help text for this question in the secondary language.')
    )
    help_lang3 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Help (tertiary)'),
        help_text=_('The help text for this question in the tertiary language.')
    )
    help_lang4 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Help (quaternary)'),
        help_text=_('The help text for this question in the quaternary language.')
    )
    help_lang5 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Help (quinary)'),
        help_text=_('The help text for this question in the quinary language.')
    )
    text_lang1 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Text (primary)'),
        help_text=_('The text for this question in the primary language.')
    )
    text_lang2 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Text (secondary)'),
        help_text=_('The text for this question in the secondary language.')
    )
    text_lang3 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Text (tertiary)'),
        help_text=_('The text for this question in the tertiary language.')
    )
    text_lang4 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Text (quaternary)'),
        help_text=_('The text for this question in the quaternary language.')
    )
    text_lang5 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Text (quinary)'),
        help_text=_('The text for this question in the quinary language.')
    )
    default_text_lang1 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Default text value (primary)'),
        help_text=_('The default text value for this question in the primary language.')
    )
    default_text_lang2 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Default text value (secondary)'),
        help_text=_('The default text value for this question in the secondary language.')
    )
    default_text_lang3 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Default text value (tertiary)'),
        help_text=_('The default text value for this question in the tertiary language.')
    )
    default_text_lang4 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Default text value (quaternary)'),
        help_text=_('The default text value for this question in the quaternary language.')
    )
    default_text_lang5 = models.TextField(
        null=True, blank=True,
        verbose_name=_('Default text value (quinary)'),
        help_text=_('The default text value for this question in the quinary language.')
    )
    default_option = models.ForeignKey(
        Option, blank=True, null=True, on_delete=models.SET_NULL,
        verbose_name=_('Default option'),
        help_text=_('The default option for this question. To be used with regular optionsets.')
    )
    default_external_id = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Default external id'),
        help_text=_('The default external id for this question. To be used with dynamic optionsets.')
    )
    verbose_name_lang1 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Name (primary)'),
        help_text=_('The name displayed for this question in the primary language.')
    )
    verbose_name_lang2 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Name (secondary)'),
        help_text=_('The name displayed for this question in the secondary language.')
    )
    verbose_name_lang3 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Name (tertiary)'),
        help_text=_('The name displayed for this question in the tertiary language.')
    )
    verbose_name_lang4 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Name (quaternary)'),
        help_text=_('The name displayed for this question in the quaternary language.')
    )
    verbose_name_lang5 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Name (quinary)'),
        help_text=_('The name displayed for this question in the quinary language.')
    )
    verbose_name_plural_lang1 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Plural name (primary)'),
        help_text=_('The plural name displayed for this question in the primary language.')
    )
    verbose_name_plural_lang2 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Plural name (secondary)'),
        help_text=_('The plural name displayed for this question in the secondary language.')
    )
    verbose_name_plural_lang3 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Plural name (tertiary)'),
        help_text=_('The plural name displayed for this question in the tertiary language.')
    )
    verbose_name_plural_lang4 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Plural name (quaternary)'),
        help_text=_('The plural name displayed for this question in the quaternary language.')
    )
    verbose_name_plural_lang5 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Plural name (quinary)'),
        help_text=_('The plural name displayed for this question in the quinary language.')
    )
    widget_type = models.CharField(
        max_length=16,
        verbose_name=_('Widget type'),
        help_text=_('Type of widget for this question.')
    )
    value_type = models.CharField(
        max_length=8, choices=VALUE_TYPE_CHOICES,
        verbose_name=_('Value type'),
        help_text=_('Type of value for this question.')
    )
    minimum = models.FloatField(
        null=True, blank=True,
        verbose_name=_('Minimum'),
        help_text=_('Minimal value for this question.')
    )
    maximum = models.FloatField(
        null=True, blank=True,
        verbose_name=_('Maximum'),
        help_text=_('Maximum value for this question.')
    )
    step = models.FloatField(
        null=True, blank=True,
        verbose_name=_('Step'),
        help_text=_('Step in which the value for this question can be incremented/decremented.')
    )
    unit = models.CharField(
        max_length=64, blank=True,
        verbose_name=_('Unit'),
        help_text=_('Unit for this question.')
    )
    width = models.IntegerField(
        null=True, blank=True,
        verbose_name=_('Width'),
        help_text=_('Width for the widget of this question (optional, full with: 12).')
    )
    optionsets = models.ManyToManyField(
        'options.OptionSet', blank=True, related_name='questions',
        verbose_name=_('Option sets'),
        help_text=_('Option sets for this question.')
    )
    conditions = models.ManyToManyField(
        Condition, blank=True, related_name='questions',
        verbose_name=_('Conditions'),
        help_text=_('List of conditions evaluated for this question.')
    )

    class Meta:
        ordering = ('questionset', 'order')
        verbose_name = _('Question')
        verbose_name_plural = _('Questions')

    def __str__(self):
        return self.path

    def save(self, *args, **kwargs):
        self.path = self.build_path(self.key, self.questionset)
        self.uri = self.build_uri(self.uri_prefix, self.path)
        super().save(*args, **kwargs)

    def copy(self, uri_prefix, key, questionset=None):
        question = copy_model(self, uri_prefix=uri_prefix, key=key, questionset=questionset or self.questionset, attribute=self.attribute)

        # copy m2m fields
        question.optionsets.set(self.optionsets.all())
        question.conditions.set(self.conditions.all())

        return question

    @property
    def parent_fields(self):
        return ('questionset', )

    @property
    def text(self):
        return self.trans('text')

    @property
    def help(self):
        return self.trans('help')

    @property
    def default_text(self):
        return self.trans('default_text')

    @property
    def verbose_name(self):
        return self.trans('verbose_name')

    @property
    def verbose_name_plural(self):
        return self.trans('verbose_name_plural')

    @property
    def is_locked(self):
        return self.locked or self.questionset.is_locked

    @property
    def is_question(self):
        return True

    @property
    def has_conditions(self):
        return self.conditions.exists()

    @classmethod
    def build_path(cls, key, questionset):
        assert key
        assert questionset
        return questionset.path + '/' + key

    @classmethod
    def build_uri(cls, uri_prefix, path):
        assert path
        return join_url(uri_prefix or settings.DEFAULT_URI_PREFIX, '/questions/', path)



from django.shortcuts import render
from django.utils.translation import gettext_lazy as _

from rdmo.core.plugins import get_plugin


def oauth_callback(request, provider_key):
    provider = get_plugin('SERVICE_PROVIDERS', provider_key)
    try:
        return provider.callback(request)
    except AssertionError:
        return render(request, 'core/error.html', {
            'title': _('Integration Error'),
            'errors': [_('Something went wrong. Please contact support.')]
        }, status=500)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion
import rdmo.core.models


class Migration(migrations.Migration):

    dependencies = [
        ('domain', '0001_initial_after_reset'),
    ]

    operations = [
        migrations.CreateModel(
            name='Catalog',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', models.DateTimeField(verbose_name='created', editable=False)),
                ('updated', models.DateTimeField(verbose_name='updated', editable=False)),
                ('order', models.IntegerField(null=True)),
                ('title_en', models.CharField(max_length=256)),
                ('title_de', models.CharField(max_length=256)),
            ],
            options={
                'ordering': ('order',),
                'verbose_name': 'Catalog',
                'verbose_name_plural': 'Catalogs',
            },
            bases=(models.Model, rdmo.core.models.TranslationMixin),
        ),
        migrations.CreateModel(
            name='QuestionEntity',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', models.DateTimeField(verbose_name='created', editable=False)),
                ('updated', models.DateTimeField(verbose_name='updated', editable=False)),
                ('order', models.IntegerField(null=True)),
                ('help_en', models.TextField(null=True, blank=True)),
                ('help_de', models.TextField(null=True, blank=True)),
            ],
            options={
                'ordering': ('subsection__section__catalog__order', 'subsection__section__order', 'subsection__order'),
                'verbose_name': 'QuestionEntity',
                'verbose_name_plural': 'QuestionEntities',
            },
            bases=(models.Model, rdmo.core.models.TranslationMixin),
        ),
        migrations.CreateModel(
            name='Section',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', models.DateTimeField(verbose_name='created', editable=False)),
                ('updated', models.DateTimeField(verbose_name='updated', editable=False)),
                ('order', models.IntegerField(null=True)),
                ('title_en', models.CharField(max_length=256)),
                ('title_de', models.CharField(max_length=256)),
                ('catalog', models.ForeignKey(related_name='sections', to='questions.Catalog', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('catalog__order', 'order'),
                'verbose_name': 'Section',
                'verbose_name_plural': 'Sections',
            },
            bases=(models.Model, rdmo.core.models.TranslationMixin),
        ),
        migrations.CreateModel(
            name='Subsection',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', models.DateTimeField(verbose_name='created', editable=False)),
                ('updated', models.DateTimeField(verbose_name='updated', editable=False)),
                ('order', models.IntegerField(null=True)),
                ('title_en', models.CharField(max_length=256)),
                ('title_de', models.CharField(max_length=256)),
                ('section', models.ForeignKey(related_name='subsections', to='questions.Section', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('section__catalog__order', 'section__order', 'order'),
                'verbose_name': 'Subsection',
                'verbose_name_plural': 'Subsections',
            },
            bases=(models.Model, rdmo.core.models.TranslationMixin),
        ),
        migrations.CreateModel(
            name='Question',
            fields=[
                ('questionentity_ptr', models.OneToOneField(parent_link=True, auto_created=True, primary_key=True, serialize=False, to='questions.QuestionEntity', on_delete=models.CASCADE)),
                ('text_en', models.TextField()),
                ('text_de', models.TextField()),
                ('widget_type', models.CharField(max_length=12, choices=[('text', 'Text'), ('textarea', 'Textarea'), ('yesno', 'Yes/No'), ('checkbox', 'Checkboxes'), ('radio', 'Radio buttons'), ('select', 'Select drop-down'), ('range', 'Range slider'), ('date', 'Date picker')])),
            ],
            options={
                'ordering': ('subsection__section__catalog__order', 'subsection__section__order', 'subsection__order'),
                'verbose_name': 'Question',
                'verbose_name_plural': 'Questions',
            },
            bases=('questions.questionentity',),
        ),
        migrations.AddField(
            model_name='questionentity',
            name='attribute_entity',
            field=models.ForeignKey(related_name='+', on_delete=django.db.models.deletion.SET_NULL, blank=True, to='domain.AttributeEntity', null=True),
        ),
        migrations.AddField(
            model_name='questionentity',
            name='subsection',
            field=models.ForeignKey(related_name='entities', to='questions.Subsection', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='question',
            name='parent_entity',
            field=models.ForeignKey(related_name='questions', blank=True, to='questions.QuestionEntity', null=True, on_delete=models.CASCADE),
        ),
    ]



# Generated by Django 2.2.18 on 2021-02-08 17:04

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('projects', '0048_meta'),
    ]

    operations = [
        migrations.CreateModel(
            name='Invite',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(blank=True, help_text='The email for this membership.', max_length=254, verbose_name='Email')),
                ('role', models.CharField(choices=[('owner', 'Owner'), ('manager', 'Manager'), ('author', 'Author'), ('guest', 'Guest')], help_text='The role for this invite.', max_length=12, verbose_name='Role')),
                ('token', models.CharField(help_text='The token for this invite.', max_length=20, verbose_name='Token')),
                ('timestamp', models.DateTimeField(help_text='The timestamp for this invite.', verbose_name='Timestamp')),
                ('project', models.ForeignKey(help_text='The project for this invite.', on_delete=django.db.models.deletion.CASCADE, related_name='invites', to='projects.Project', verbose_name='Project')),
                ('user', models.ForeignKey(help_text='The user for this membership.', null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Invite',
                'verbose_name_plural': 'Invites',
                'ordering': ('timestamp',),
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion

import rdmo.core.models

class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AttributeEntity',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=256)),
                ('full_title', models.CharField(max_length=256)),
                ('description', models.TextField(null=True, blank=True)),
                ('uri', models.URLField(null=True, blank=True)),
                ('is_collection', models.BooleanField(default=False)),
            ],
            options={
                'verbose_name': 'AttributeEntity',
                'verbose_name_plural': 'AttributeEntities',
            },
        ),
        migrations.CreateModel(
            name='Condition',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('relation', models.CharField(max_length=8, choices=[('eq', 'equal (==)'), ('neq', 'not equal (!=)'), ('contains', 'contains'), ('gt', 'greater than (>)'), ('gte', 'greater than or equal (>=)'), ('lt', 'lesser than (<)'), ('lte', 'lesser than or equal (<=)')])),
                ('target_text', models.CharField(max_length=256, null=True, blank=True)),
            ],
            options={
                'verbose_name': 'Condition',
                'verbose_name_plural': 'Conditions',
            },
        ),
        migrations.CreateModel(
            name='Option',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('order', models.IntegerField(null=True)),
                ('text_en', models.CharField(max_length=256)),
                ('text_de', models.CharField(max_length=256)),
                ('additional_input', models.BooleanField()),
            ],
            options={
                'ordering': ('attribute', 'order'),
                'verbose_name': 'Option',
                'verbose_name_plural': 'Options',
            },
            bases=(models.Model, rdmo.core.models.TranslationMixin),
        ),
        migrations.CreateModel(
            name='Range',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('minimum', models.FloatField()),
                ('maximum', models.FloatField()),
                ('step', models.FloatField()),
            ],
            options={
                'ordering': ('attribute',),
                'verbose_name': 'Range',
                'verbose_name_plural': 'Ranges',
            },
            bases=(models.Model, rdmo.core.models.TranslationMixin),
        ),
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('attributeentity_ptr', models.OneToOneField(parent_link=True, auto_created=True, primary_key=True, serialize=False, to='domain.AttributeEntity', on_delete=models.CASCADE)),
                ('value_type', models.CharField(max_length=8, choices=[('text', 'Text'), ('integer', 'Integer'), ('float', 'Float'), ('boolean', 'Boolean'), ('datetime', 'Datetime'), ('options', 'Options')])),
                ('unit', models.CharField(max_length=64, null=True, blank=True)),
            ],
            options={
                'verbose_name': 'Attribute',
                'verbose_name_plural': 'Attributes',
            },
            bases=('domain.attributeentity',),
        ),
        migrations.AddField(
            model_name='condition',
            name='target_option',
            field=models.ForeignKey(related_name='+', on_delete=django.db.models.deletion.SET_NULL, blank=True, to='domain.Option', null=True),
        ),
        migrations.AddField(
            model_name='attributeentity',
            name='parent_entity',
            field=models.ForeignKey(related_name='children', blank=True, to='domain.AttributeEntity', help_text='optional', null=True, on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='range',
            name='attribute',
            field=models.OneToOneField(to='domain.Attribute', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='option',
            name='attribute',
            field=models.ForeignKey(related_name='options', to='domain.Attribute', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='condition',
            name='attribute',
            field=models.ForeignKey(related_name='conditions', to='domain.Attribute', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='condition',
            name='source_attribute',
            field=models.ForeignKey(related_name='+', on_delete=django.db.models.deletion.SET_NULL, blank=True, to='domain.Attribute', null=True),
        ),
    ]



from django.urls import re_path

from ..views import (IntegrationCreateView, IntegrationDeleteView,
                     IntegrationUpdateView, IntegrationWebhookView,
                     InviteDeleteView, IssueDetailView, IssueSendView,
                     IssueUpdateView, MembershipCreateView,
                     MembershipDeleteView, MembershipUpdateView,
                     ProjectAnswersExportView, ProjectAnswersView,
                     ProjectCancelView, ProjectCreateImportView,
                     ProjectCreateView, ProjectDeleteView, ProjectDetailView,
                     ProjectErrorView, ProjectExportView, ProjectJoinView,
                     ProjectLeaveView, ProjectQuestionsView, ProjectsView,
                     ProjectUpdateCatalogView, ProjectUpdateImportView,
                     ProjectUpdateInformationView, ProjectUpdateParentView,
                     ProjectUpdateTasksView, ProjectUpdateView,
                     ProjectUpdateViewsView, ProjectViewExportView,
                     ProjectViewView, SiteProjectsView, SnapshotCreateView,
                     SnapshotRollbackView, SnapshotUpdateView)

urlpatterns = [
    re_path(r'^$', ProjectsView.as_view(), name='projects'),
    re_path(r'^all/$', SiteProjectsView.as_view(), name='site_projects'),

    re_path(r'^create/$', ProjectCreateView.as_view(), name='project_create'),
    re_path(r'^join/(?P<token>.+)/$', ProjectJoinView.as_view(), name='project_join'),
    re_path(r'^cancel/(?P<token>.+)/$', ProjectCancelView.as_view(), name='project_cancel'),
    re_path(r'^import/$', ProjectCreateImportView.as_view(), name='project_create_import'),

    re_path(r'^(?P<pk>[0-9]+)/$', ProjectDetailView.as_view(), name='project'),
    re_path(r'^(?P<pk>[0-9]+)/update/$', ProjectUpdateView.as_view(), name='project_update'),
    re_path(r'^(?P<pk>[0-9]+)/update/information/$', ProjectUpdateInformationView.as_view(), name='project_update_information'),
    re_path(r'^(?P<pk>[0-9]+)/update/catalog/$', ProjectUpdateCatalogView.as_view(), name='project_update_catalog'),
    re_path(r'^(?P<pk>[0-9]+)/update/parent/$', ProjectUpdateParentView.as_view(), name='project_update_parent'),
    re_path(r'^(?P<pk>[0-9]+)/update/tasks/$', ProjectUpdateTasksView.as_view(), name='project_update_tasks'),
    re_path(r'^(?P<pk>[0-9]+)/update/views/$', ProjectUpdateViewsView.as_view(), name='project_update_views'),
    re_path(r'^(?P<pk>[0-9]+)/delete/$', ProjectDeleteView.as_view(), name='project_delete'),
    re_path(r'^(?P<pk>[0-9]+)/leave/$', ProjectLeaveView.as_view(), name='project_leave'),
    re_path(r'^(?P<pk>[0-9]+)/export/(?P<format>[a-z]+)/$', ProjectExportView.as_view(), name='project_export'),
    re_path(r'^(?P<pk>[0-9]+)/import/$', ProjectUpdateImportView.as_view(), name='project_update_import'),

    re_path(r'^(?P<project_id>[0-9]+)/memberships/create/$', MembershipCreateView.as_view(), name='membership_create'),
    re_path(r'^(?P<project_id>[0-9]+)/memberships/(?P<pk>[0-9]+)/update/$', MembershipUpdateView.as_view(), name='membership_update'),
    re_path(r'^(?P<project_id>[0-9]+)/memberships/(?P<pk>[0-9]+)/delete/$', MembershipDeleteView.as_view(), name='membership_delete'),
    re_path(r'^(?P<project_id>[0-9]+)/invite/(?P<pk>[0-9]+)/delete/$', InviteDeleteView.as_view(), name='invite_delete'),

    re_path(r'^(?P<project_id>[0-9]+)/integrations/create/(?P<provider_key>[a-z]+)/$', IntegrationCreateView.as_view(), name='integration_create'),
    re_path(r'^(?P<project_id>[0-9]+)/integrations/(?P<pk>[0-9]+)/update/$', IntegrationUpdateView.as_view(), name='integration_update'),
    re_path(r'^(?P<project_id>[0-9]+)/integrations/(?P<pk>[0-9]+)/delete/$', IntegrationDeleteView.as_view(), name='integration_delete'),
    re_path(r'^(?P<project_id>[0-9]+)/integrations/(?P<pk>[0-9]+)/webhook/$', IntegrationWebhookView.as_view(), name='integration_webhook'),

    re_path(r'^(?P<project_id>[0-9]+)/issues/(?P<pk>[0-9]+)/$', IssueDetailView.as_view(), name='issue'),
    re_path(r'^(?P<project_id>[0-9]+)/issues/(?P<pk>[0-9]+)/update/$', IssueUpdateView.as_view(), name='issue_update'),
    re_path(r'^(?P<project_id>[0-9]+)/issues/(?P<pk>[0-9]+)/send/$', IssueSendView.as_view(), name='issue_send'),

    re_path(r'^(?P<project_id>[0-9]+)/snapshots/create/$', SnapshotCreateView.as_view(), name='snapshot_create'),
    re_path(r'^(?P<project_id>[0-9]+)/snapshots/(?P<pk>[0-9]+)/update/$', SnapshotUpdateView.as_view(), name='snapshot_update'),
    re_path(r'^(?P<project_id>[0-9]+)/snapshots/(?P<pk>[0-9]+)/rollback/$', SnapshotRollbackView.as_view(), name='snapshot_rollback'),

    re_path(r'^(?P<pk>[0-9]+)/answers/$', ProjectAnswersView.as_view(), name='project_answers'),
    re_path(r'^(?P<pk>[0-9]+)/answers/export/(?P<format>[a-z]+)/$', ProjectAnswersExportView.as_view(), name='project_answers_export'),

    re_path(r'^(?P<pk>[0-9]+)/snapshots/(?P<snapshot_id>[0-9]+)/answers/$', ProjectAnswersView.as_view(), name='project_answers'),
    re_path(r'^(?P<pk>[0-9]+)/snapshots/(?P<snapshot_id>[0-9]+)/answers/export/(?P<format>[a-z]+)/$', ProjectAnswersExportView.as_view(), name='project_answers_export'),

    re_path(r'^(?P<pk>[0-9]+)/views/(?P<view_id>[0-9]+)/$', ProjectViewView.as_view(), name='project_view'),
    re_path(r'^(?P<pk>[0-9]+)/views/(?P<view_id>[0-9]+)/export/(?P<format>[a-z]+)/$', ProjectViewExportView.as_view(), name='project_view_export'),

    re_path(r'^(?P<pk>[0-9]+)/snapshots/(?P<snapshot_id>[0-9]+)/views/(?P<view_id>[0-9]+)/$', ProjectViewView.as_view(), name='project_view'),
    re_path(r'^(?P<pk>[0-9]+)/snapshots/(?P<snapshot_id>[0-9]+)/views/(?P<view_id>[0-9]+)/export/(?P<format>[a-z]+)/$', ProjectViewExportView.as_view(), name='project_view_export'),

    re_path(r'^(?P<pk>[0-9]+)/questions/', ProjectQuestionsView.as_view(), name='project_questions'),
    re_path(r'^(?P<pk>[0-9]+)/error/', ProjectErrorView.as_view(), name='project_error'),
]



from django.utils.translation import gettext_lazy as _

DEBUG = False

INSTALLED_APPS = [
    # django modules
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    # rdmo modules
    'rdmo',
    'rdmo.core',
    'rdmo.overlays',
    'rdmo.accounts',
    'rdmo.services',
    'rdmo.domain',
    'rdmo.options',
    'rdmo.conditions',
    'rdmo.questions',
    'rdmo.tasks',
    'rdmo.views',
    'rdmo.projects',
    'rdmo.management',
    # 3rd party modules
    'rest_framework',
    'rest_framework.authtoken',
    'widget_tweaks',
    'markdown',
    'compressor',
    'django_cleanup',
    'django_extensions',
    'django_filters',
    'mathfilters',
    'mptt',
    'rules',
    # openapi specification tools
    'rest_framework_swagger'
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.contrib.sites.middleware.CurrentSiteMiddleware'
]

ROOT_URLCONF = 'config.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django_settings_export.settings_export',
            ],
        },
    },
]

COMPRESS_PRECOMPILERS = (
    ('text/x-scss', 'django_libsass.SassCompiler'),
)

WSGI_APPLICATION = 'config.wsgi.application'

AUTHENTICATION_BACKENDS = [
    'rules.permissions.ObjectPermissionBackend',
    'django.contrib.auth.backends.ModelBackend'
]

MULTISITE = False

PROFILE_UPDATE = True
PROFILE_DELETE = True

ACCOUNT = False
ACCOUNT_SIGNUP = False
ACCOUNT_TERMS_OF_USE = False

SOCIALACCOUNT = False

SHIBBOLETH = False
SHIBBOLETH_LOGOUT_URL = '/Shibboleth.sso/Logout'

ACCOUNT_SIGNUP_FORM_CLASS = 'rdmo.accounts.forms.SignupForm'
ACCOUNT_USER_DISPLAY = 'rdmo.accounts.utils.get_full_name'
ACCOUNT_EMAIL_REQUIRED = True
ACCOUNT_ACTIVATION_DAYS = 7
ACCOUNT_EMAIL_CONFIRMATION_EXPIRE_DAYS = 7
ACCOUNT_EMAIL_VERIFICATION = 'optional'
ACCOUNT_LOGIN_ON_EMAIL_CONFIRMATION = False
ACCOUNT_USERNAME_MIN_LENGTH = 4
ACCOUNT_PASSWORD_MIN_LENGTH = 4
ACCOUNT_EMAIL_MAX_LENGTH = 190

ACCOUNT_ADAPTER = 'rdmo.accounts.adapter.AccountAdapter'

SOCIALACCOUNT_ADAPTER = 'rdmo.accounts.adapter.SocialAccountAdapter'
SOCIALACCOUNT_SIGNUP = False
SOCIALACCOUNT_AUTO_SIGNUP = False

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'Europe/Berlin'

LANGUAGES = (
    ('en', _('English')),
    ('de', _('German')),
)

USE_I18N = True

USE_L10N = True

USE_TZ = True

LOGIN_URL = '/account/login/'
LOGIN_REDIRECT_URL = '/'
LOGOUT_URL = '/account/logout/'

MEDIA_URL = '/media/'
STATIC_URL = '/static/'

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'compressor.finders.CompressorFinder',
)

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite3'
    }
}

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'rdmo_default'
    }
}

REST_FRAMEWORK = {
    'DEFAULT_SCHEMA_CLASS': 'rest_framework.schemas.coreapi.AutoSchema',
    'UNICODE_JSON': False,
    'DEFAULT_VERSIONING_CLASS': 'rest_framework.versioning.NamespaceVersioning',
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.TokenAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    ),
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework.renderers.JSONRenderer',
    )
}

SETTINGS_EXPORT = [
    'LOGIN_URL',
    'LOGOUT_URL',
    'ACCOUNT',
    'ACCOUNT_SIGNUP',
    'ACCOUNT_TERMS_OF_USE',
    'SOCIALACCOUNT',
    'PROFILE_UPDATE',
    'PROFILE_DELETE',
    'SHIBBOLETH',
    'MULTISITE',
    'EXPORT_FORMATS',
    'PROJECT_ISSUES',
    'PROJECT_VIEWS',
    'PROJECT_EXPORTS',
    'PROJECT_IMPORTS',
    'PROJECT_SEND_ISSUE',
    'PROJECT_QUESTIONS_AUTOSAVE',
    'NESTED_PROJECTS'
]

SETTINGS_API = [
    'PROJECT_QUESTIONS_AUTOSAVE',
    'PROJECT_QUESTIONS_CYCLE_SETS',
    'DEFAULT_URI_PREFIX'
]

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'
DEFAULT_FROM_EMAIL = 'info@example.com'

EMAIL_RECIPIENTS_CHOICES = []
EMAIL_RECIPIENTS_INPUT = False

USER_API = True

OVERLAYS = {
    'projects': [
        'projects-table',
        'create-project',
        'import-project'
    ],
    'project': [
        'project-questions',
        'project-catalog',
        'project-issues',
        'project-views',
        'project-memberships',
        'project-snapshots',
        'export-project',
        'import-project'
    ],
    'issue_send': [
        'issue-message',
        'issue-attachments'
    ]
}

EXPORT_FORMATS = (
    ('pdf', _('PDF')),
    ('rtf', _('Rich Text Format')),
    ('odt', _('Open Office')),
    ('docx', _('Microsoft Office')),
    ('html', _('HTML')),
    ('markdown', _('Markdown')),
    ('mediawiki', _('mediawiki')),
    ('tex', _('LaTeX'))
)

EXPORT_REFERENCE_ODT_VIEWS = {}
EXPORT_REFERENCE_DOCX_VIEWS = {}
EXPORT_REFERENCE_ODT = None
EXPORT_REFERENCE_DOCX = None

EXPORT_PANDOC_ARGS = {
    'pdf': ['-V', 'geometry:margin=1in', '--pdf-engine=xelatex'],
    'rtf': ['--standalone']
}

PROJECT_ISSUES = True

PROJECT_VIEWS = True

PROJECT_EXPORTS = [
    ('xml', _('RDMO XML'), 'rdmo.projects.exports.RDMOXMLExport'),
    ('csvcomma', _('CSV comma separated'), 'rdmo.projects.exports.CSVCommaExport'),
    ('csvsemicolon', _('CSV semicolon separated'), 'rdmo.projects.exports.CSVSemicolonExport')
]

PROJECT_IMPORTS = [
    ('xml', _('RDMO XML'), 'rdmo.projects.imports.RDMOXMLImport'),
]

PROJECT_QUESTIONS_AUTOSAVE = False

PROJECT_QUESTIONS_CYCLE_SETS = False

PROJECT_FILE_QUOTA = '10Mb'

PROJECT_SEND_ISSUE = False

PROJECT_INVITE_TIMEOUT = None

PROJECT_SEND_INVITE = True

NESTED_PROJECTS = True

SERVICE_PROVIDERS = []

OPTIONSET_PROVIDERS = []

QUESTIONS_WIDGETS = [
    ('text', _('Text'), 'rdmo.projects.widgets.TextWidget'),
    ('textarea', _('Textarea'), 'rdmo.projects.widgets.TextareaWidget'),
    ('yesno', _('Yes/No'), 'rdmo.projects.widgets.YesnoWidget'),
    ('checkbox', _('Checkboxes'), 'rdmo.projects.widgets.CheckboxWidget'),
    ('radio', _('Radio buttons'), 'rdmo.projects.widgets.RadioWidget'),
    ('select', _('Select drop-down'), 'rdmo.projects.widgets.SelectWidget'),
    ('autocomplete', _('Autocomplete'), 'rdmo.projects.widgets.AutocompleteWidget'),
    ('range', _('Range slider'), 'rdmo.projects.widgets.RangeWidget'),
    ('date', _('Date picker'), 'rdmo.projects.widgets.DateWidget'),
    ('file', _('File upload'), 'rdmo.projects.widgets.FileWidget')
]

DEFAULT_URI_PREFIX = 'http://example.com/terms'

VENDOR_CDN = True

VENDOR = {
    'jquery': {
        'url': 'https://code.jquery.com/',
        'js': [
            {
                'path': 'jquery-3.4.1.min.js',
                'sri': 'sha384-vk5WoKIaW/vJyUAd9n/wmopsmNhiy+L2Z+SBxGYnUkunIxVxAv/UtMOhba/xskxh',
            }
        ]
    },
    'bootstrap': {
        'url': 'https://maxcdn.bootstrapcdn.com/bootstrap/3.4.1/',
        'js': [
            {
                'path': 'js/bootstrap.min.js',
                'sri': 'sha384-aJ21OjlMXNL5UyIl/XNwTMqvzeRMZH2w8c5cRVpzpU8Y5bApTppSuUkhZXN0VxHd',
            }
        ],
        'css': [
            {
                'path': 'css/bootstrap.min.css',
                'sri': 'sha384-HSMxcRTRxnN+Bdg0JdbxYKrThecOKuH5zCYotlSAcp1+c8xmyTe9GYg1l9a69psu',
            }
        ]
    },
    'bootstrap-datepicker': {
        'url': 'https://cdnjs.cloudflare.com/ajax/libs/bootstrap-datepicker/1.7.1/',
        'css': [
            {
                'path': 'css/bootstrap-datepicker.min.css'
            }
        ],
        'js': [
            {
                'path': 'js/bootstrap-datepicker.min.js'
            }
        ]
    },
    'font-awesome': {
        'url': 'https://maxcdn.bootstrapcdn.com/font-awesome/4.7.0/',
        'css': [
            {
                'path': 'css/font-awesome.min.css'
            }
        ],
        'font': [
            {
                'path': 'fonts/fontawesome-webfont.eot'
            },
            {
                'path': 'fonts/fontawesome-webfont.woff2'
            },
            {
                'path': 'fonts/fontawesome-webfont.woff'
            },
            {
                'path': 'fonts/fontawesome-webfont.ttf'
            },
            {
                'path': 'fonts/fontawesome-webfont.svg'
            }
        ]
    },
    'angular': {
        'url': 'https://ajax.googleapis.com/ajax/libs/angularjs/1.5.8/',
        'js': [
            {
                'path': 'angular.min.js'
            },
            {
                'path': 'angular-resource.min.js'
            }
        ]
    },
    'select2': {
        'url': 'https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.6-rc.0/',
        'js': [
            {
                'path': 'js/select2.min.js',
                'sri': 'sha256-HNkbndPiWM5EIRgahc3hWiuGD6CtwFgMfEU0o3zeabo='
            }
        ],
        'css': [
            {
                'path': 'css/select2.min.css',
                'sri': 'sha256-EQA4j7+ZbrewCQvwJzNmVxiKMwGRspXMGgt7I6AAiqs='
            }
        ]
    },
    'select2-bootstrap-theme': {
        'url': 'https://cdnjs.cloudflare.com/ajax/libs/select2-bootstrap-theme/0.1.0-beta.10/',
        'css': [
            {
                'path': 'select2-bootstrap.min.css',
                'sri': 'sha256-nbyata2PJRjImhByQzik2ot6gSHSU4Cqdz5bNYL2zcU='
            }
        ]
    },
    'moment': {
        'url': 'https://cdnjs.cloudflare.com/ajax/libs/moment.js/2.18.1/',
        'js': [
            {
                'path': 'moment.min.js',
                'sri': 'sha256-1hjUhpc44NwiNg8OwMu2QzJXhD8kcj+sJA3aCQZoUjg='
            }
        ]
    },
    'codemirror': {
        'url': 'https://cdnjs.cloudflare.com/ajax/libs/codemirror/5.29.0/',
        'js': [
            {
                'path': 'codemirror.min.js',
                'sri': 'sha256-0LRLvWWVXwt0eH0/Bzd0PHICg/bSMDIe5sXgaDSpZaA='
            },
            {
                'path': 'addon/mode/overlay.min.js',
                'sri': 'sha256-ffWkw3Pn4ieMygm1vwdRKcMtBJ6E6kuBi8GlVVPXWEs='
            },
            {
                'path': 'mode/django/django.min.js',
                'sri': 'sha256-6hO1TjC+3W73p+kXnCqcHVjfRa4KMdG7hvWencnu0XM='
            }
        ],
        'css': [
            {
                'path': 'codemirror.min.css',
                'sri': 'sha256-wluO/w4cnorJpS0JmcdTSYzwdb5E6u045qa4Ervfb1k='
            }
        ]
    },
    'fuse': {
        'url': 'https://cdnjs.cloudflare.com/ajax/libs/fuse.js/3.4.6/',
        'js': [
            {
                'path': 'fuse.min.js',
                'sri': 'sha512-FwWaT/y9ajd/+J06KL9Fko1jELonJNHMUTR4nGP9MSIq4ZdU2w9/OiLxn16p/zEOZkryHi3wKYsnWPuADD328Q=='
            }
        ]
    }
}

# necessary since django 3.2, explicitly set primary key type to avaoid warnings
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'



import logging

from django.conf import settings
from django.db.models import Prefetch
from django.shortcuts import redirect
from django.views.generic import DetailView

from rdmo.core.constants import VALUE_TYPE_FILE
from rdmo.core.utils import render_to_format
from rdmo.core.views import ObjectPermissionMixin
from rdmo.questions.models import QuestionSet, Question
from rdmo.views.utils import ProjectWrapper

from ..models import Project, Snapshot
from ..utils import get_value_path
logger = logging.getLogger(__name__)


class ProjectAnswersView(ObjectPermissionMixin, DetailView):
    model = Project
    queryset = Project.objects.prefetch_related(
        Prefetch('catalog__sections__questionsets', queryset=QuestionSet.objects.select_related('attribute')),
        Prefetch('catalog__sections__questionsets__questions', queryset=Question.objects.select_related('attribute', 'questionset')),
        Prefetch('catalog__sections__questionsets__questionsets', queryset=QuestionSet.objects.select_related('attribute')),
        Prefetch('catalog__sections__questionsets__questionsets__questions', queryset=Question.objects.select_related('attribute', 'questionset')),
    )

    permission_required = 'projects.view_project_object'
    template_name = 'projects/project_answers.html'
    no_catalog_error_template = 'projects/project_error_no_catalog.html'

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        if self.object.catalog is None:
            return redirect('project_error', pk=self.object.pk)
        else:
            context = self.get_context_data(object=self.object)
            return self.render_to_response(context)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        try:
            context['current_snapshot'] = context['project'].snapshots.get(pk=self.kwargs.get('snapshot_id'))
        except Snapshot.DoesNotExist:
            context['current_snapshot'] = None

        # collect values with files, remove double files and order them.
        context['attachments'] = context['project'].values.filter(snapshot=context['current_snapshot']) \
                                                          .filter(value_type=VALUE_TYPE_FILE) \
                                                          .order_by('file')

        context.update({
            'project_wrapper': ProjectWrapper(context['project'], context['current_snapshot']),
            'snapshots': list(context['project'].snapshots.values('id', 'title')),
            'export_formats': settings.EXPORT_FORMATS
        })

        return context


class ProjectAnswersExportView(ObjectPermissionMixin, DetailView):
    model = Project
    queryset = Project.objects.prefetch_related(
        Prefetch('catalog__sections__questionsets', queryset=QuestionSet.objects.select_related('attribute')),
        Prefetch('catalog__sections__questionsets__questions', queryset=Question.objects.select_related('attribute', 'questionset')),
        Prefetch('catalog__sections__questionsets__questionsets', queryset=QuestionSet.objects.select_related('attribute')),
        Prefetch('catalog__sections__questionsets__questionsets__questions', queryset=Question.objects.select_related('attribute', 'questionset')),
    )
    permission_required = 'projects.view_project_object'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        try:
            context['current_snapshot'] = context['project'].snapshots.get(pk=self.kwargs.get('snapshot_id'))
        except Snapshot.DoesNotExist:
            context['current_snapshot'] = None

        context.update({
            'project_wrapper': ProjectWrapper(context['project'], context['current_snapshot']),
            'title': context['project'].title,
            'format': self.kwargs.get('format'),
            'resource_path': get_value_path(context['project'], context['current_snapshot'])
        })

        return context

    def render_to_response(self, context, **response_kwargs):
        return render_to_format(self.request, context['format'], context['title'], 'projects/project_answers_export.html', context)



from rdmo.conditions.imports import import_condition
from rdmo.core.constants import PERMISSIONS
from rdmo.domain.imports import fetch_attribute_parents, import_attribute
from rdmo.options.imports import (fetch_option_parents, import_option,
                                  import_optionset)
from rdmo.questions.imports import (fetch_question_parents,
                                    fetch_questionset_parents,
                                    fetch_section_parents, import_catalog,
                                    import_question, import_questionset,
                                    import_section)
from rdmo.tasks.imports import import_task
from rdmo.views.imports import import_view


def check_permissions(elements, user):
    element_types = set([element.get('type') for element in elements])

    permissions = []
    for element_type in element_types:
        permissions += PERMISSIONS[element_type]

    return user.has_perms(permissions)


def get_parent_uri(element_uri, parent_uri, parents, uris):
    if element_uri in parents and parents[element_uri]:
        # parent was explicitely selected
        return parents[element_uri]
    elif parent_uri in uris:
        # parent uri was changed during the import
        return uris[parent_uri]
    else:
        # needs to be False, to use parent uri in file
        return False


def import_elements(elements, parents={}, save={}):
    instances = []
    uris = {}

    for element in elements:
        element_type = element.get('type')
        save_element = save.get(element.get('uri'))

        # step 1: get create model for elements
        if element_type == 'condition':
            instance = import_condition(element, save=save_element)

        elif element_type == 'attribute':
            parent_uri = get_parent_uri(element.get('uri'), element.get('parent'), parents, uris)
            instance = import_attribute(element, parent_uri=parent_uri, save=save_element)

        elif element_type == 'optionset':
            instance = import_optionset(element, save=save_element)

        elif element_type == 'option':
            optionset_uri = get_parent_uri(element.get('uri'), element.get('optionset'), parents, uris)
            instance = import_option(element, optionset_uri=optionset_uri, save=save_element)

        elif element_type == 'catalog':
            instance = import_catalog(element, save=save_element)

        elif element_type == 'section':
            catalog_uri = get_parent_uri(element.get('uri'), element.get('catalog'), parents, uris)
            instance = import_section(element, catalog_uri=catalog_uri, save=save_element)

        elif element_type == 'questionset':
            section_uri = get_parent_uri(element.get('uri'), element.get('section'), parents, uris)
            questionset_uri = get_parent_uri(element.get('uri'), element.get('questionset'), {}, uris)
            instance = import_questionset(element, section_uri=section_uri, questionset_uri=questionset_uri, save=save_element)

        elif element_type == 'question':
            questionset_uri = get_parent_uri(element.get('uri'), element.get('questionset'), parents, uris)
            instance = import_question(element, questionset_uri=questionset_uri, save=save_element)

        elif element_type == 'task':
            instance = import_task(element, save=save_element)

        elif element_type == 'view':
            instance = import_view(element, save=save_element)

        else:
            instance = None

        # step 2: fetch available parents
        if instance:
            if not save:
                if element_type == 'attribute':
                    instance.parents = fetch_attribute_parents(instances)

                elif element.get('type') == 'option':
                    instance.parents = fetch_option_parents(instances)

                elif element.get('type') == 'section':
                    instance.parents = fetch_section_parents(instances)

                elif element.get('type') == 'questionset':
                    instance.parents = fetch_questionset_parents(instances)

                elif element.get('type') == 'question':
                    instance.parents = fetch_question_parents(instances)

                # check if a missing element was already imported
                for uri in instance.missing:
                    if uri in [instance.uri for instance in instances]:
                        instance.missing[uri]['in_file'] = True

            # append the instance to the list of instances
            instances.append(instance)

            if element.get('uri') != instance.uri:
                uris[element.get('uri')] = instance.uri

    return instances



from django.urls import include, path, re_path

from rdmo.core.views import i18n_switcher

urlpatterns = [
    path('account/', include('rdmo.accounts.urls')),
    path('conditions/', include('rdmo.conditions.urls')),
    path('domain/', include('rdmo.domain.urls')),
    path('management/', include('rdmo.management.urls')),
    path('options/', include('rdmo.options.urls')),
    path('overlays/', include('rdmo.overlays.urls')),
    path('projects/', include('rdmo.projects.urls')),
    path('questions/', include('rdmo.questions.urls')),
    path('services/', include('rdmo.services.urls')),
    path('tasks/', include('rdmo.tasks.urls')),
    path('views/', include('rdmo.views.urls')),

    re_path(r'^i18n/([a-z]{2})/$', i18n_switcher, name='i18n_switcher'),
]



from ..models import View


def test_view_str(db):
    instances = View.objects.all()
    for instance in instances:
        assert str(instance)


def test_view_clean(db):
    instances = View.objects.all()
    for instance in instances:
        instance.clean()


def test_view_copy(db):
    instances = View.objects.all()
    for instance in instances:
        new_uri_prefix = instance.uri_prefix + '-'
        new_key = instance.key + '-'
        new_instance = instance.copy(new_uri_prefix, new_key)
        assert new_instance.uri_prefix == new_uri_prefix
        assert new_instance.key == new_key
        assert list(new_instance.catalogs.values('id')) == list(new_instance.catalogs.values('id'))
        assert list(new_instance.sites.values('id')) == list(new_instance.sites.values('id'))
        assert list(new_instance.groups.values('id')) == list(new_instance.groups.values('id'))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-11-29 11:47
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('projects', '0022_move_attribute_to_attributeentity'),
    ]

    operations = [
        migrations.AlterField(
            model_name='value',
            name='value_type',
            field=models.CharField(choices=[('text', 'Text'), ('url', 'URL'), ('integer', 'Integer'), ('float', 'Float'), ('boolean', 'Boolean'), ('datetime', 'Datetime'), ('option', 'Option')], help_text='Type of this value.', max_length=8, verbose_name='Value type'),
        ),
    ]



# Generated by Django 2.2.16 on 2021-01-21 10:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('domain', '0046_parent_cascade'),
    ]

    operations = [
        migrations.AddField(
            model_name='attribute',
            name='locked',
            field=models.BooleanField(default=False, help_text="Designates whether this attribute (and it\'s descendants) can be changed.", verbose_name='Locked'),
        ),
    ]



import xml.etree.ElementTree as et

import pytest
from django.urls import reverse

from ..models import Question

users = (
    ('editor', 'editor'),
    ('reviewer', 'reviewer'),
    ('user', 'user'),
    ('api', 'api'),
    ('anonymous', None),
)

status_map = {
    'list': {
        'editor': 200, 'reviewer': 200, 'api': 200, 'user': 403, 'anonymous': 401
    },
    'detail': {
        'editor': 200, 'reviewer': 200, 'api': 200, 'user': 403, 'anonymous': 401
    },
    'create': {
        'editor': 201, 'reviewer': 403, 'api': 201, 'user': 403, 'anonymous': 401
    },
    'update': {
        'editor': 200, 'reviewer': 403, 'api': 200, 'user': 403, 'anonymous': 401
    },
    'delete': {
        'editor': 204, 'reviewer': 403, 'api': 204, 'user': 403, 'anonymous': 401
    }
}

urlnames = {
    'list': 'v1-questions:question-list',
    'nested': 'v1-questions:question-nested',
    'index': 'v1-questions:question-index',
    'export': 'v1-questions:question-export',
    'detail': 'v1-questions:question-detail',
    'detail_export': 'v1-questions:question-detail-export',
    'copy': 'v1-questions:question-copy'
}


@pytest.mark.parametrize('username,password', users)
def test_list(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse(urlnames['list'])
    response = client.get(url)
    assert response.status_code == status_map['list'][username], response.json()


@pytest.mark.parametrize('username,password', users)
def test_index(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse(urlnames['index'])
    response = client.get(url)
    assert response.status_code == status_map['list'][username], response.json()


@pytest.mark.parametrize('username,password', users)
def test_export(db, client, username, password):
    client.login(username=username, password=password)

    url = reverse(urlnames['export'])
    response = client.get(url)
    assert response.status_code == status_map['list'][username], response.content

    if response.status_code == 200:
        root = et.fromstring(response.content)
        assert root.tag == 'rdmo'
        for child in root:
            assert child.tag in ['question']


@pytest.mark.parametrize('username,password', users)
def test_detail(db, client, username, password):
    client.login(username=username, password=password)
    instances = Question.objects.all()

    for instance in instances:
        url = reverse(urlnames['detail'], args=[instance.pk])
        response = client.get(url)
        assert response.status_code == status_map['detail'][username], response.json()


@pytest.mark.parametrize('username,password', users)
def test_nested(db, client, username, password):
    client.login(username=username, password=password)
    instances = Question.objects.all()

    for instance in instances:
        url = reverse(urlnames['nested'], args=[instance.pk])
        response = client.get(url)
        assert response.status_code == status_map['detail'][username], response.json()


@pytest.mark.parametrize('username,password', users)
def test_create(db, client, username, password):
    client.login(username=username, password=password)
    instances = Question.objects.all()

    for instance in instances:
        url = reverse(urlnames['list'])
        data = {
            'uri_prefix': instance.uri_prefix,
            'key': '%s_new_%s' % (instance.key, username),
            'comment': instance.comment or '',
            'attribute': instance.attribute.pk if instance.attribute else '',
            'questionset': instance.questionset.pk,
            'is_collection': instance.is_collection,
            'order': instance.order,
            'help_en': instance.help_lang1 or '',
            'help_de': instance.help_lang2 or '',
            'text_en': instance.text_lang1 or '',
            'text_de': instance.text_lang2 or '',
            'verbose_name_en': instance.verbose_name_lang1 or '',
            'verbose_name_de': instance.verbose_name_lang2 or '',
            'verbose_name_plural_en': instance.verbose_name_plural_lang1 or '',
            'verbose_name_plural_de': instance.verbose_name_plural_lang2 or '',
            'widget_type': instance.widget_type,
            'value_type': instance.value_type,
            'minimum': instance.minimum or '',
            'maximum': instance.maximum or '',
            'step': instance.step or '',
            'unit': instance.unit or '',
            'optionsets': [optionset.pk for optionset in instance.optionsets.all()],
            'conditions': [condition.pk for condition in instance.conditions.all()]
        }
        response = client.post(url, data, content_type='application/json')
        assert response.status_code == status_map['create'][username], response.json()


@pytest.mark.parametrize('username,password', users)
def test_update(db, client, username, password):
    client.login(username=username, password=password)
    instances = Question.objects.all()

    for instance in instances:
        url = reverse(urlnames['detail'], args=[instance.pk])
        data = {
            'uri_prefix': instance.uri_prefix,
            'key': instance.key,
            'comment': instance.comment,
            'attribute': instance.attribute.pk if instance.attribute else None,
            'questionset': instance.questionset.pk,
            'is_collection': instance.is_collection,
            'order': instance.order,
            'help_en': instance.help_lang1,
            'help_de': instance.help_lang2,
            'text_en': instance.text_lang1,
            'text_de': instance.text_lang2,
            'verbose_name_en': instance.verbose_name_lang1,
            'verbose_name_de': instance.verbose_name_lang2,
            'verbose_name_plural_en': instance.verbose_name_plural_lang1,
            'verbose_name_plural_de': instance.verbose_name_plural_lang2,
            'widget_type': instance.widget_type,
            'value_type': instance.value_type,
            'minimum': instance.minimum,
            'maximum': instance.maximum,
            'step': instance.step,
            'unit': instance.unit,
            'optionsets': [optionset.pk for optionset in instance.optionsets.all()],
            'conditions': [condition.pk for condition in instance.conditions.all()],
        }
        response = client.put(url, data, content_type='application/json')
        assert response.status_code == status_map['update'][username], response.json()


@pytest.mark.parametrize('username,password', users)
def test_delete(db, client, username, password):
    client.login(username=username, password=password)
    instances = Question.objects.all()

    for instance in instances:
        url = reverse(urlnames['detail'], args=[instance.pk])
        response = client.delete(url)
        assert response.status_code == status_map['delete'][username], response.json()


@pytest.mark.parametrize('username,password', users)
def test_detail_export(db, client, username, password):
    client.login(username=username, password=password)
    instances = Question.objects.all()

    for instance in instances:
        url = reverse(urlnames['detail_export'], args=[instance.pk])
        response = client.get(url)
        assert response.status_code == status_map['list'][username], response.content

        if response.status_code == 200:
            root = et.fromstring(response.content)
            assert root.tag == 'rdmo'
            for child in root:
                assert child.tag in ['question']


@pytest.mark.parametrize('username,password', users)
def test_copy(db, client, username, password):
    client.login(username=username, password=password)
    instances = Question.objects.all()

    for instance in instances:
        url = reverse(urlnames['copy'], args=[instance.pk])
        data = {
            'uri_prefix': instance.uri_prefix + '-',
            'key': instance.key + '-',
            'questionset': instance.questionset.id
        }
        response = client.put(url, data, content_type='application/json')
        assert response.status_code == status_map['create'][username], response.json()


@pytest.mark.parametrize('username,password', users)
def test_copy_wrong(db, client, username, password):
    client.login(username=username, password=password)
    instance = Question.objects.first()

    url = reverse(urlnames['copy'], args=[instance.pk])
    data = {
        'uri_prefix': instance.uri_prefix,
        'key': instance.key,
        'questionset': instance.questionset.id
    }
    response = client.put(url, data, content_type='application/json')

    if status_map['create'][username] == 201:
        assert response.status_code == 400, response.json()
    else:
        assert response.status_code == status_map['create'][username], response.json()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('domain', '0002_additional_input'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='condition',
            name='attribute',
        ),
        migrations.AddField(
            model_name='condition',
            name='attribute_entity',
            field=models.ForeignKey(related_name='conditions', default=None, to='domain.AttributeEntity', on_delete=models.CASCADE),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('domain', '0004_verbosename'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='attributeentity',
            options={'ordering': ('full_title',), 'verbose_name': 'AttributeEntity', 'verbose_name_plural': 'AttributeEntities'},
        ),
    ]



from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.sites.models import Site
from django.db import models
from django.utils.translation import gettext_lazy as _
from rdmo.conditions.models import Condition
from rdmo.core.models import TranslationMixin
from rdmo.core.utils import copy_model, join_url
from rdmo.domain.models import Attribute
from rdmo.questions.models import Catalog

from .managers import TaskManager


class Task(TranslationMixin, models.Model):

    objects = TaskManager()

    uri = models.URLField(
        max_length=640, blank=True,
        verbose_name=_('URI'),
        help_text=_('The Uniform Resource Identifier of this task (auto-generated).')
    )
    uri_prefix = models.URLField(
        max_length=256,
        verbose_name=_('URI Prefix'),
        help_text=_('The prefix for the URI of this task.')
    )
    key = models.SlugField(
        max_length=128, blank=True,
        verbose_name=_('Key'),
        help_text=_('The internal identifier of this task.')
    )
    comment = models.TextField(
        blank=True,
        verbose_name=_('Comment'),
        help_text=_('Additional internal information about this task.')
    )
    locked = models.BooleanField(
        default=False,
        verbose_name=_('Locked'),
        help_text=_('Designates whether this task can be changed.')
    )
    catalogs = models.ManyToManyField(
        Catalog, blank=True,
        verbose_name=_('Catalogs'),
        help_text=_('The catalogs this task can be used with. An empty list implies that this task can be used with every catalog.')
    )
    sites = models.ManyToManyField(
        Site, blank=True,
        verbose_name=_('Sites'),
        help_text=_('The sites this task belongs to (in a multi site setup).')
    )
    groups = models.ManyToManyField(
        Group, blank=True,
        verbose_name=_('Group'),
        help_text=_('The groups for which this task is active.')
    )
    title_lang1 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Title (primary)'),
        help_text=_('The title for this task in the primary language.')
    )
    title_lang2 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Title (secondary)'),
        help_text=_('The title for this task in the secondary language.')
    )
    title_lang3 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Title (tertiary)'),
        help_text=_('The title for this task in the tertiary language.')
    )
    title_lang4 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Title (quaternary)'),
        help_text=_('The title for this task in the quaternary language.')
    )
    title_lang5 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Title (quinary)'),
        help_text=_('The title for this task in the quinary language.')
    )
    text_lang1 = models.TextField(
        blank=True,
        verbose_name=_('Text (primary)'),
        help_text=_('The text for this task in the primary language.')
    )
    text_lang2 = models.TextField(
        blank=True,
        verbose_name=_('Text (secondary)'),
        help_text=_('The text for this task in the secondary language.')
    )
    text_lang3 = models.TextField(
        blank=True,
        verbose_name=_('Text (tertiary)'),
        help_text=_('The text for this task in the tertiary language.')
    )
    text_lang4 = models.TextField(
        blank=True,
        verbose_name=_('Text (quaternary)'),
        help_text=_('The text for this task in the quaternary language.')
    )
    text_lang5 = models.TextField(
        blank=True,
        verbose_name=_('Text (quinary)'),
        help_text=_('The text for this task in the quinary language.')
    )
    start_attribute = models.ForeignKey(
        Attribute, blank=True, null=True, on_delete=models.SET_NULL, related_name='tasks_as_start',
        verbose_name=_('Start date attribute'),
        help_text=_('The attribute that is setting the start date for this task.')
    )
    end_attribute = models.ForeignKey(
        Attribute, blank=True, null=True, on_delete=models.SET_NULL, related_name='tasks_as_end',
        verbose_name=_('End date attribute'),
        help_text=_('The attribute that is setting the end date for this task (optional, if no end date attribute is given, the start date attribute sets also the end date).')
    )
    days_before = models.IntegerField(
        blank=True, null=True,
        verbose_name=_('Days before'),
        help_text=_('Additional days before the start date.')
    )
    days_after = models.IntegerField(
        blank=True, null=True,
        verbose_name=_('Days after'),
        help_text=_('Additional days after the end date.')
    )
    conditions = models.ManyToManyField(
        Condition, blank=True, related_name='tasks',
        verbose_name=_('Conditions'),
        help_text=_('The list of conditions evaluated for this task.')
    )
    available = models.BooleanField(
        default=True,
        verbose_name=_('Available'),
        help_text=_('Designates whether this task is generally available for projects.')
    )

    class Meta:
        ordering = ('uri',)
        verbose_name = _('Task')
        verbose_name_plural = _('Tasks')

    def __str__(self):
        return self.key

    def save(self, *args, **kwargs):
        self.uri = self.build_uri(self.uri_prefix, self.key)
        super().save(*args, **kwargs)

    def copy(self, uri_prefix, key):
        task = copy_model(self, uri_prefix=uri_prefix, key=key, start_attribute=self.start_attribute, end_attribute=self.end_attribute)

        # add m2m fields
        task.catalogs.set(self.catalogs.all())
        task.sites.set(self.sites.all())
        task.groups.set(self.groups.all())
        task.conditions.set(self.conditions.all())

        return task

    @property
    def title(self):
        return self.trans('title')

    @property
    def text(self):
        return self.trans('text')

    @property
    def has_conditions(self):
        return bool(self.conditions.all())

    @property
    def is_locked(self):
        return self.locked

    @classmethod
    def build_uri(cls, uri_prefix, key):
        assert key
        return join_url(uri_prefix or settings.DEFAULT_URI_PREFIX, '/tasks/', key)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-02-03 13:07
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('options', '0009_data_migration'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='option',
            options={'ordering': ('optionset__order', 'order'), 'verbose_name': 'Option', 'verbose_name_plural': 'Options'},
        ),
        migrations.AlterField(
            model_name='option',
            name='path',
            field=models.SlugField(blank=True, help_text='The path part of the URI for this option (auto-generated).', max_length=512, null=True, verbose_name='Path'),
        ),
    ]



from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


def run_data_migration(apps, schema_editor):
    Project = apps.get_model('projects', 'Project')
    Task = apps.get_model('tasks', 'Task')
    View = apps.get_model('views', 'View')

    for project in Project.objects.all():
        # add all tasks to project
        tasks = Task.objects.filter(sites=settings.SITE_ID)
        for task in tasks:
            project.tasks.add(task)

        # add all views to project
        views = View.objects.filter(sites=settings.SITE_ID).filter(models.Q(catalogs=None) | models.Q(catalogs=project.catalog))
        for view in views:
            project.views.add(view)


class Migration(migrations.Migration):

    dependencies = [
        ('projects', '0031_project_tasks'),
        ('tasks', '0028_data_migration'),
        ('views', '0023_data_migration')
    ]

    operations = [
        migrations.RunPython(run_data_migration),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-04-04 14:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('options', '0011_permissions'),
    ]

    operations = [
        migrations.AlterField(
            model_name='option',
            name='comment',
            field=models.TextField(blank=True, help_text='Additional internal information about this option.', null=True, verbose_name='Comment'),
        ),
        migrations.AlterField(
            model_name='option',
            name='key',
            field=models.SlugField(blank=True, help_text='The internal identifier of this option.', max_length=128, null=True, verbose_name='Key'),
        ),
        migrations.AlterField(
            model_name='option',
            name='order',
            field=models.IntegerField(default=0, help_text='Position in lists.', verbose_name='Order'),
        ),
        migrations.AlterField(
            model_name='optionset',
            name='comment',
            field=models.TextField(blank=True, help_text='Additional internal information about this option set.', null=True, verbose_name='Comment'),
        ),
        migrations.AlterField(
            model_name='optionset',
            name='key',
            field=models.SlugField(blank=True, help_text='The internal identifier of this option set.', max_length=128, null=True, verbose_name='Key'),
        ),
    ]



from django.conf import settings
from django.contrib.sites.models import Site
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from rdmo.core.models import TranslationMixin


class AdditionalField(models.Model, TranslationMixin):

    TYPE_CHOICES = (
        ('text', 'Text'),
        ('textarea', 'Textarea'),
    )

    key = models.SlugField()
    type = models.CharField(max_length=11, choices=TYPE_CHOICES)

    text_lang1 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Text (primary)'),
        help_text=_('The text for this additional field in the primary language.')
    )
    text_lang2 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Text (secondary)'),
        help_text=_('The text for this additional field in the secondary language.')
    )
    text_lang3 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Text (tertiary)'),
        help_text=_('The text for this additional field in the tertiary language.')
    )
    text_lang4 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Text (quaternary)'),
        help_text=_('The text for this additional field in the quaternary language.')
    )
    text_lang5 = models.CharField(
        max_length=256, blank=True,
        verbose_name=_('Text (quinary)'),
        help_text=_('The text for this additional field in the quinary language.')
    )
    help_lang1 = models.TextField(
        blank=True,
        verbose_name=_('Help (primary)'),
        help_text=_('The help text to be displayed next to the input element in the primary language.')
    )
    help_lang2 = models.TextField(
        blank=True,
        verbose_name=_('Help (secondary)'),
        help_text=_('The help text to be displayed next to the input element in the secondary language.')
    )
    help_lang3 = models.TextField(
        blank=True,
        verbose_name=_('Help (tertiary)'),
        help_text=_('The help text to be displayed next to the input element in the tertiary language.')
    )
    help_lang4 = models.TextField(
        blank=True,
        verbose_name=_('Help (quaternary)'),
        help_text=_('The help text to be displayed next to the input element in the quaternary language.')
    )
    help_lang5 = models.TextField(
        blank=True,
        verbose_name=_('Help (quinary)'),
        help_text=_('The help text to be displayed next to the input element in the quinary language.')
    )
    required = models.BooleanField(
        verbose_name=_('Required'),
        help_text=_('Designates whether this additional field is required.')
    )

    class Meta:
        ordering = ('key',)
        verbose_name = _('Additional field')
        verbose_name_plural = _('Additional fields')

    def __str__(self):
        return self.text

    @property
    def text(self):
        return self.trans('text')

    @property
    def help(self):
        return self.trans('help')


class AdditionalFieldValue(models.Model):

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='additional_values')
    field = models.ForeignKey(AdditionalField, on_delete=models.CASCADE, related_name='+')
    value = models.CharField(max_length=256)

    class Meta:
        ordering = ('user', 'field')

        verbose_name = _('Additional field value')
        verbose_name_plural = _('Additional field values')

    def __str__(self):
        return self.user.username + '/' + self.field.key


class ConsentFieldValue(models.Model):

    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    consent = models.BooleanField(
        default=False,
        help_text='Designates whether the user has agreed to the terms of use.',
        verbose_name='Consent'
    )

    class Meta:
        ordering = ('user', )
        verbose_name = _('Consent field value')
        verbose_name_plural = _('Consent field values')

    def __str__(self):
        return self.user.username


class Role(models.Model):

    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    member = models.ManyToManyField(
        Site, related_name='members', blank=True,
        verbose_name=_('Member'),
        help_text=_('The sites for which this user is a member.')
    )
    manager = models.ManyToManyField(
        Site, related_name='managers', blank=True,
        verbose_name=_('Manager'),
        help_text=_('The sites for which this user is manager.')
    )

    class Meta:
        ordering = ('user', )
        verbose_name = _('Role')
        verbose_name_plural = _('Roles')

    def __str__(self):
        return self.user.username


@receiver(post_save, sender=settings.AUTH_USER_MODEL)
def post_save_user(sender, **kwargs):
    if not kwargs.get('raw', False):
        user = kwargs['instance']
        current_site = Site.objects.get_current()

        try:
            role = user.role
        except Role.DoesNotExist:
            role = Role(user=user)
            role.save()

        if current_site not in role.member.all():
            role.member.add(current_site)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-04-21 13:56
from __future__ import unicode_literals

import django.contrib.auth.models
from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0008_alter_user_username_max_length'),
        ('accounts', '0008_related_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProxyUser',
            fields=[
            ],
            options={
                'default_permissions': (),
                'proxy': True,
                'permissions': (('view_user', 'Can view user'),),
            },
            bases=(settings.AUTH_USER_MODEL,),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
    ]
